#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for the CF reader."""

import datetime as dt
import warnings

import numpy as np
import pytest
import xarray as xr
from pyresample import AreaDefinition, SwathDefinition

from satpy import Scene
from satpy.dataset.dataid import WavelengthRange
from satpy.readers.satpy_cf_nc import SatpyCFFileHandler

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


def _create_test_netcdf(filename, resolution=742):
    size = 2
    if resolution == 371:
        size = 4
    data_visir = np.array(np.arange(1.0, size * size + 1)).reshape(size, size)
    lat = 33.0 * data_visir
    lon = -13.0 * data_visir

    lat = xr.DataArray(lat,
                       dims=("y", "x"),
                       attrs={"name": "lat",
                              "standard_name": "latitude",
                              "modifiers": np.array([])})
    lon = xr.DataArray(lon,
                       dims=("y", "x"),
                       attrs={"name": "lon",
                              "standard_name": "longitude",
                              "modifiers": np.array([])})

    solar_zenith_angle_i = xr.DataArray(data_visir,
                                        dims=("y", "x"),
                                        attrs={"name": "solar_zenith_angle",
                                               "coordinates": "lat lon",
                                               "resolution": resolution})

    scene = Scene()
    scene.attrs["sensor"] = ["viirs"]
    scene_dict = {
        "lat": lat,
        "lon": lon,
        "solar_zenith_angle": solar_zenith_angle_i
    }

    tstart = dt.datetime(2019, 4, 1, 12, 0)
    tend = dt.datetime(2019, 4, 1, 12, 15)
    common_attrs = {
        "start_time": tstart,
        "end_time": tend,
        "platform_name": "NOAA 20",
        "orbit_number": 99999
    }

    for key in scene_dict:
        scene[key] = scene_dict[key]
        if key != "swath_data":
            scene[key].attrs.update(common_attrs)
    scene.save_datasets(writer="cf",
                        filename=filename,
                        engine="h5netcdf",
                        flatten_attrs=True,
                        pretty=True)
    return filename


@pytest.fixture(scope="module")
def area():
    """Get area definition."""
    area_extent = (339045.5577, 4365586.6063, 1068143.527, 4803645.4685)
    proj_dict = {"a": 6378169.0, "b": 6356583.8, "h": 35785831.0,
                 "lon_0": 0.0, "proj": "geos", "units": "m"}
    area = AreaDefinition("test",
                          "test",
                          "test",
                          proj_dict,
                          2,
                          2,
                          area_extent)
    return area


@pytest.fixture(scope="module")
def common_attrs(area):
    """Get common dataset attributes."""
    return {
        "start_time": dt.datetime(2019, 4, 1, 12, 0, 0, 123456),
        "end_time": dt.datetime(2019, 4, 1, 12, 15),
        "platform_name": "tirosn",
        "orbit_number": 99999,
        "area": area,
        "my_timestamp": dt.datetime(2000, 1, 1)
    }


@pytest.fixture(scope="module")
def xy_coords(area):
    """Get projection coordinates."""
    x, y = area.get_proj_coords()
    y = y[:, 0]
    x = x[0, :]
    return x, y


@pytest.fixture(scope="module")
def vis006(xy_coords, common_attrs):
    """Get VIS006 dataset."""
    x, y = xy_coords
    attrs = {
        "name": "image0",
        "id_tag": "ch_r06",
        "coordinates": "lat lon",
        "resolution": 1000,
        "calibration": "reflectance",
        "wavelength": WavelengthRange(min=0.58, central=0.63, max=0.68, unit="µm"),
        "orbital_parameters": {
          "projection_longitude": 1,
          "projection_latitude": 1,
          "projection_altitude": 1,
          "satellite_nominal_longitude": 1,
          "satellite_nominal_latitude": 1,
          "satellite_actual_longitude": 1,
          "satellite_actual_latitude": 1,
          "satellite_actual_altitude": 1,
          "nadir_longitude": 1,
          "nadir_latitude": 1,
          "only_in_1": False
        },
        "time_parameters": {
          "nominal_start_time": common_attrs["start_time"],
          "nominal_end_time": common_attrs["end_time"]
        }
    }
    coords = {"y": y, "x": x, "acq_time": ("y", [1, 2])}
    vis006 = xr.DataArray(np.array([[1.0, 2.0], [3.0, 4.0]]),
                          dims=("y", "x"),
                          coords=coords,
                          attrs=attrs)
    return vis006


@pytest.fixture(scope="module")
def ir_108(xy_coords):
    """Get IR_108 dataset."""
    x, y = xy_coords
    coords = {"y": y, "x": x, "acq_time": ("y", [1, 2])}
    attrs = {"name": "image1", "id_tag": "ch_tb11", "coordinates": "lat lon"}
    ir_108 = xr.DataArray(np.array([[1.0, 2.0], [3.0, 4.0]]),
                          dims=("y", "x"),
                          coords=coords,
                          attrs=attrs)
    return ir_108


@pytest.fixture(scope="module")
def qual_flags(xy_coords):
    """Get quality flags."""
    qual_data = np.array(
        [[1, 2, 3, 4, 5, 6, 7],
         [1, 2, 3, 4, 5, 6, 7]],
        dtype=np.int8)
    x, y = xy_coords
    z = [1, 2, 3, 4, 5, 6, 7]
    coords = {"y": y, "z": z, "acq_time": ("y", [1, 2])}
    qual_f = xr.DataArray(qual_data,
                          dims=("y", "z"),
                          coords=coords,
                          attrs={"name": "qual_flags",
                                 "id_tag": "qual_flags"})
    return qual_f


@pytest.fixture(scope="module")
def lonlats(xy_coords):
    """Get longitudes and latitudes."""
    x, y = xy_coords
    lat = 33.0 * np.array([[1, 2], [3, 4]])
    lon = -13.0 * np.array([[1, 2], [3, 4]])
    attrs = {"name": "lat",
             "standard_name": "latitude",
             "modifiers": np.array([])}
    dims = ("y", "x")
    coords = {"y": y, "x": x}
    lat = xr.DataArray(lat, dims=dims, coords=coords, attrs=attrs)
    lon = xr.DataArray(lon, dims=dims, coords=coords, attrs=attrs)
    return lon, lat


@pytest.fixture(scope="module")
def prefix_data(xy_coords, area):
    """Get dataset whose name should be prefixed."""
    x, y = xy_coords
    attrs = {"name": "1",
             "id_tag": "ch_r06",
             "coordinates": "lat lon",
             "resolution": 1000,
             "calibration": "reflectance",
             "wavelength": WavelengthRange(min=0.58, central=0.63, max=0.68, unit="µm"),
             "area": area}
    prefix_data = xr.DataArray(np.array([[1.0, 2.0], [3.0, 4.0]]),
                               dims=("y", "x"),
                               coords={"y": y, "x": x},
                               attrs=attrs)
    return prefix_data


@pytest.fixture(scope="module")
def swath_data(prefix_data, lonlats):
    """Get swath data."""
    lon, lat = lonlats
    area = SwathDefinition(lons=lon, lats=lat)
    swath_data = prefix_data.copy()
    swath_data.attrs.update({"name": "swath_data", "area": area})
    return swath_data


@pytest.fixture(scope="module")
def datasets(vis006, ir_108, qual_flags, lonlats, prefix_data, swath_data):
    """Get datasets belonging to the scene."""
    lon, lat = lonlats
    return {"image0": vis006,
            "image1": ir_108,
            "swath_data": swath_data,
            "1": prefix_data,
            "lat": lat,
            "lon": lon,
            "qual_flags": qual_flags}


@pytest.fixture(scope="module")
def cf_scene(datasets, common_attrs):
    """Create a cf scene."""
    scene = Scene()
    scene.attrs["sensor"] = ["avhrr-1", "avhrr-2", "avhrr-3"]
    for key in datasets:
        scene[key] = datasets[key]
        if key != "swath_data":
            scene[key].attrs.update(common_attrs)
    return scene


@pytest.fixture
def nc_filename(tmp_path):
    """Create an nc filename for viirs m band."""
    now = dt.datetime.now(dt.timezone.utc)
    filename = f"testingcfwriter{now:%Y%j%H%M%S}-viirs-mband-20201007075915-20201007080744.nc"
    return str(tmp_path / filename)


@pytest.fixture
def nc_filename_i(tmp_path):
    """Create an nc filename for viirs i band."""
    now = dt.datetime.now(dt.timezone.utc)
    filename = f"testingcfwriter{now:%Y%j%H%M%S}-viirs-iband-20201007075915-20201007080744.nc"
    return str(tmp_path / filename)


class TestCFReader:
    """Test case for CF reader."""

    def test_write_and_read_with_area_definition(self, cf_scene, nc_filename):
        """Save a dataset with an area definition to file with cf_writer and read the data again."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename,
                               engine="h5netcdf",
                               flatten_attrs=True,
                               pretty=True)
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["image0", "image1", "lat"])
        np.testing.assert_array_equal(scn_["image0"].data, cf_scene["image0"].data)
        np.testing.assert_array_equal(scn_["lat"].data, cf_scene["lat"].data)  # lat loaded as dataset
        np.testing.assert_array_equal(scn_["image0"].coords["lon"], cf_scene["lon"].data)  # lon loded as coord
        assert isinstance(scn_["image0"].attrs["wavelength"], WavelengthRange)
        expected_area = cf_scene["image0"].attrs["area"]
        actual_area = scn_["image0"].attrs["area"]
        assert pytest.approx(expected_area.area_extent, 0.000001) == actual_area.area_extent
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",
                                    message=r"You will likely lose important projection information",
                                    category=UserWarning)
            assert expected_area.proj_dict == actual_area.proj_dict
        assert expected_area.shape == actual_area.shape
        assert expected_area.area_id == actual_area.area_id
        assert expected_area.description == actual_area.description

    def test_write_and_read_with_swath_definition(self, cf_scene, nc_filename):
        """Save a dataset with a swath definition to file with cf_writer and read the data again."""
        with warnings.catch_warnings():
            # Filter out warning about missing lon/lat DataArray coordinates
            warnings.filterwarnings("ignore", category=UserWarning, message=r"Coordinate .* referenced")
            cf_scene.save_datasets(writer="cf",
                                filename=nc_filename,
                                engine="h5netcdf",
                                flatten_attrs=True,
                                pretty=True,
                                datasets=["swath_data"])
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["swath_data"])
        expected_area = cf_scene["swath_data"].attrs["area"]
        actual_area = scn_["swath_data"].attrs["area"]
        assert expected_area.shape == actual_area.shape
        np.testing.assert_array_equal(expected_area.lons.data, actual_area.lons.data)
        np.testing.assert_array_equal(expected_area.lats.data, actual_area.lats.data)

    def test_fix_modifier_attr(self):
        """Check that fix modifier can handle empty list as modifier attribute."""
        reader = SatpyCFFileHandler("filename",
                                    {},
                                    {"filetype": "info"})
        ds_info = {"modifiers": []}
        reader.fix_modifier_attr(ds_info)
        assert ds_info["modifiers"] == ()

    def test_read_prefixed_channels(self, cf_scene, nc_filename):
        """Check channels starting with digit is prefixed and read back correctly."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename,
                               engine="netcdf4",
                               flatten_attrs=True,
                               pretty=True)
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["1"])
        np.testing.assert_array_equal(scn_["1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["1"].coords["lon"], cf_scene["lon"].data)  # lon loaded as coord

        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename], reader_kwargs={})
        scn_.load(["1"])
        np.testing.assert_array_equal(scn_["1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["1"].coords["lon"], cf_scene["lon"].data)  # lon loaded as coord

        # Check that variables starting with a digit is written to filename variable prefixed
        with xr.open_dataset(nc_filename) as ds_disk:
            np.testing.assert_array_equal(ds_disk["CHANNEL_1"].data, cf_scene["1"].data)

    def test_read_prefixed_channels_include_orig_name(self, cf_scene, nc_filename):
        """Check channels starting with digit and includeed orig name is prefixed and read back correctly."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename,
                               engine="netcdf4",
                               flatten_attrs=True,
                               pretty=True,
                               include_orig_name=True)
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["1"])
        np.testing.assert_array_equal(scn_["1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["1"].coords["lon"], cf_scene["lon"].data)  # lon loaded as coord

        assert scn_["1"].attrs["original_name"] == "1"

        # Check that variables starting with a digit is written to filename variable prefixed
        with xr.open_dataset(nc_filename) as ds_disk:
            np.testing.assert_array_equal(ds_disk["CHANNEL_1"].data, cf_scene["1"].data)

    def test_read_prefixed_channels_by_user(self, cf_scene, nc_filename):
        """Check channels starting with digit is prefixed by user and read back correctly."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename,
                               engine="netcdf4",
                               flatten_attrs=True,
                               pretty=True,
                               numeric_name_prefix="USER")
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename], reader_kwargs={"numeric_name_prefix": "USER"})
        scn_.load(["1"])
        np.testing.assert_array_equal(scn_["1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["1"].coords["lon"], cf_scene["lon"].data)  # lon loded as coord

        # Check that variables starting with a digit is written to filename variable prefixed
        with xr.open_dataset(nc_filename) as ds_disk:
            np.testing.assert_array_equal(ds_disk["USER1"].data, cf_scene["1"].data)

    def test_read_prefixed_channels_by_user2(self, cf_scene, nc_filename):
        """Check channels starting with digit is prefixed by user when saving and read back correctly without prefix."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename,
                               engine="netcdf4",
                               flatten_attrs=True,
                               pretty=True,
                               include_orig_name=False,
                               numeric_name_prefix="USER")
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["USER1"])
        np.testing.assert_array_equal(scn_["USER1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["USER1"].coords["lon"], cf_scene["lon"].data)  # lon loded as coord

    def test_read_prefixed_channels_by_user_include_prefix(self, cf_scene, nc_filename):
        """Check channels starting with digit is prefixed by user and include original name when saving."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename,
                               engine="netcdf4",
                               flatten_attrs=True,
                               pretty=True,
                               include_orig_name=True,
                               numeric_name_prefix="USER")
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["1"])
        np.testing.assert_array_equal(scn_["1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["1"].coords["lon"], cf_scene["lon"].data)  # lon loded as coord

    def test_read_prefixed_channels_by_user_no_prefix(self, cf_scene, nc_filename):
        """Check channels starting with digit is not prefixed by user."""
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning, message=".*starts with a digit.*")
            cf_scene.save_datasets(writer="cf",
                                   filename=nc_filename,
                                   engine="netcdf4",
                                   flatten_attrs=True,
                                   pretty=True,
                                   numeric_name_prefix="")
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["1"])
        np.testing.assert_array_equal(scn_["1"].data, cf_scene["1"].data)
        np.testing.assert_array_equal(scn_["1"].coords["lon"], cf_scene["lon"].data)  # lon loded as coord

    def test_decoding_of_dict_type_attributes(self, cf_scene, nc_filename):
        """Test decoding of dict type attributes."""
        cf_scene.save_datasets(writer="cf",
                               filename=nc_filename)
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename])
        scn_.load(["image0"])
        for attr_name in ["orbital_parameters", "time_parameters"]:
            orig_attrs = cf_scene["image0"].attrs[attr_name]
            new_attrs = scn_["image0"].attrs[attr_name]
            assert new_attrs == orig_attrs

    def test_decoding_of_timestamps(self, cf_scene, nc_filename):
        """Test decoding of timestamps."""
        cf_scene.save_datasets(writer="cf", filename=nc_filename)
        scn = Scene(reader="satpy_cf_nc", filenames=[nc_filename])
        scn.load(["image0"])
        expected = cf_scene["image0"].attrs["my_timestamp"]
        assert scn["image0"].attrs["my_timestamp"] == expected

    def test_write_and_read_from_two_files(self, nc_filename, nc_filename_i):
        """Save two datasets with different resolution and read the solar_zenith_angle again."""
        _create_test_netcdf(nc_filename, resolution=742)
        _create_test_netcdf(nc_filename_i, resolution=371)
        scn_ = Scene(reader="satpy_cf_nc",
                     filenames=[nc_filename, nc_filename_i])
        scn_.load(["solar_zenith_angle"], resolution=742)
        assert scn_["solar_zenith_angle"].attrs["resolution"] == 742
        scn_.unload()
        scn_.load(["solar_zenith_angle"], resolution=371)
        assert scn_["solar_zenith_angle"].attrs["resolution"] == 371

    def test_dataid_attrs_equal_matching_dataset(self, cf_scene, nc_filename):
        """Check that get_dataset returns valid dataset when keys matches."""
        from satpy.dataset.dataid import DataID, default_id_keys_config
        _create_test_netcdf(nc_filename, resolution=742)
        reader = SatpyCFFileHandler(nc_filename, {}, {"filetype": "info"})
        ds_id = DataID(default_id_keys_config, name="solar_zenith_angle", resolution=742, modifiers=())
        res = reader.get_dataset(ds_id, {})
        assert res.attrs["resolution"] == 742

    def test_dataid_attrs_equal_not_matching_dataset(self, cf_scene, nc_filename):
        """Check that get_dataset returns None when key(s) are not matching."""
        from satpy.dataset.dataid import DataID, default_id_keys_config
        _create_test_netcdf(nc_filename, resolution=742)
        reader = SatpyCFFileHandler(nc_filename, {}, {"filetype": "info"})
        not_existing_resolution = 9999999
        ds_id = DataID(default_id_keys_config, name="solar_zenith_angle", resolution=not_existing_resolution,
                       modifiers=())
        assert reader.get_dataset(ds_id, {}) is None

    def test_dataid_attrs_equal_contains_not_matching_key(self, cf_scene, nc_filename):
        """Check that get_dataset returns valid dataset when dataid have key(s) not existing in data."""
        from satpy.dataset.dataid import DataID, default_id_keys_config
        _create_test_netcdf(nc_filename, resolution=742)
        reader = SatpyCFFileHandler(nc_filename, {}, {"filetype": "info"})
        ds_id = DataID(default_id_keys_config, name="solar_zenith_angle", resolution=742,
                       modifiers=(), calibration="counts")
        res = reader.get_dataset(ds_id, {})
        assert res.attrs["resolution"] == 742



# Copyright (c) 2024 - 2025 Pytroll Developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for the Arctic Weather Satellite (AWS) MWR level-1c data.

MWR = Microwave Radiometer, onboard AWS and EPS-Sterna

Sample data provided by ESA September 27, 2024.


Example:
--------
Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene
    from glob import glob

    filenames = glob("data/W_XX-OHB-Stockholm,SAT,AWS1-MWR-1C-RAD_C_OHB_*20240913204851_*.nc")

    scn = Scene(filenames=filenames, reader='aws1_mwr_l1c_nc')

    composites = ['mw183_humidity']
    dataset_names = composites + ['1']

    scn.load(dataset_names)
    print(scn['1'])
    scn.show('mw183_humidity')

"""


from satpy.readers.mwr_l1b import MWR_CHANNEL_NAMES, AWS_EPS_Sterna_BaseFileHandler, mask_and_scale

NAVIGATION_DATASET_NAMES = ["satellite_zenith_angle",
                            "solar_azimuth_angle",
                            "solar_zenith_angle",
                            "satellite_azimuth_angle",
                            "surface_type",
                            "terrain_elevation",
                            "longitude",
                            "latitude"]

class AWS_MWR_L1CFile(AWS_EPS_Sterna_BaseFileHandler):
    """Class implementing the AWS L1c Filehandler.

    This class implements the ESA Arctic Weather Satellite (AWS) Level-1b
    NetCDF reader. It is designed to be used through the :class:`Scene <satpy.scene.Scene>`
    class using the :mod:`Scene.load <satpy.scene.Scene.load>` method with the reader
    ``"aws_l1c_nc"``.

    """
    def __init__(self, filename, filename_info, filetype_info, auto_maskandscale=True):
        """Initialize the handler."""
        super().__init__(filename, filename_info, filetype_info, auto_maskandscale)
        self.filename_info = filename_info

    @property
    def sensor(self):
        """Get the sensor name."""
        # This should have been self["/attr/instrument"]
        # But the sensor name is currently incorrect in the ESA level-1b files
        return "mwr"

    def get_dataset(self, dataset_id, dataset_info):
        """Get the data."""
        if dataset_id["name"] in MWR_CHANNEL_NAMES:
            data_array = self._get_channel_data(dataset_id, dataset_info)

        elif dataset_id["name"] in NAVIGATION_DATASET_NAMES:
            data_array = self._get_navigation_data(dataset_id, dataset_info)
        else:
            raise NotImplementedError(f"Dataset {dataset_id['name']} not available or not supported yet!")

        data_array = mask_and_scale(data_array)
        if dataset_id["name"] == "longitude":
            data_array = data_array.where(data_array <= 180, data_array - 360)

        data_array.attrs.update(dataset_info)

        data_array.attrs["platform_name"] = self.platform_name
        data_array.attrs["sensor"] = self.sensor
        return data_array


    def _get_navigation_data(self, dataset_id, dataset_info):
        """Get the navigation (geolocation) data."""
        geo_data = self[dataset_info["file_key"]]
        geo_data = geo_data.rename({"n_fovs": "x", "n_scans": "y"})
        return geo_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019, 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Test the eps l1b format."""

import os
from contextlib import suppress
from tempfile import mkstemp
from unittest import TestCase, mock

import numpy as np
import pytest
import xarray as xr

import satpy
from satpy._config import get_config_path
from satpy.readers import eps_l1b as eps
from satpy.tests.utils import make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - caplog

grh_dtype = np.dtype([("record_class", "|i1"),
                      ("INSTRUMENT_GROUP", "|i1"),
                      ("RECORD_SUBCLASS", "|i1"),
                      ("RECORD_SUBCLASS_VERSION", "|i1"),
                      ("RECORD_SIZE", ">u4"),
                      ("RECORD_START_TIME", "S6"),
                      ("RECORD_STOP_TIME", "S6")])


def create_sections(structure):
    """Create file sections."""
    sections = {}
    format_fn = get_config_path("eps_avhrrl1b_6.5.xml")
    form = eps.XMLFormat(format_fn)
    for count, (rec_class, sub_class) in structure:
        try:
            the_dtype = form.dtype((rec_class, sub_class))
        except KeyError:
            continue
        item_size = the_dtype.itemsize + grh_dtype.itemsize
        the_dtype = np.dtype(grh_dtype.descr + the_dtype.descr)
        item = np.zeros(count, the_dtype)
        item["record_class"] = eps.record_class.index(rec_class)
        item["RECORD_SUBCLASS"] = sub_class
        item["RECORD_SIZE"] = item_size

        sections[(rec_class, sub_class)] = item
    return sections


class BaseTestCaseEPSL1B(TestCase):
    """Base class for EPS l1b test case."""

    def _create_structure(self):
        structure = [(1, ("mphr", 0)), (1, ("sphr", 0)), (11, ("ipr", 0)),
                     (1, ("geadr", 1)), (1, ("geadr", 2)), (1, ("geadr", 3)),
                     (1, ("geadr", 4)), (1, ("geadr", 5)), (1, ("geadr", 6)),
                     (1, ("geadr", 7)), (1, ("giadr", 1)), (1, ("giadr", 2)),
                     (1, ("veadr", 1)), (self.scan_lines, ("mdr", 2))]
        sections = create_sections(structure)
        return sections


class TestEPSL1B(BaseTestCaseEPSL1B):
    """Test the filehandler."""

    def setUp(self):
        """Set up the tests."""
        # ipr is not present in the xml format ?
        self.scan_lines = 1080
        self.earth_views = 2048

        sections = self._create_structure()
        sections[("mphr", 0)]["TOTAL_MDR"] = (b"TOTAL_MDR                     =   " +
                                              bytes(str(self.scan_lines), encoding="ascii") +
                                              b"\n")
        sections[("mphr", 0)]["SPACECRAFT_ID"] = b"SPACECRAFT_ID                 = M03\n"
        sections[("mphr", 0)]["INSTRUMENT_ID"] = b"INSTRUMENT_ID                 = AVHR\n"
        sections[("sphr", 0)]["EARTH_VIEWS_PER_SCANLINE"] = (b"EARTH_VIEWS_PER_SCANLINE      =  " +
                                                             bytes(str(self.earth_views), encoding="ascii") +
                                                             b"\n")
        sections[("sphr", 0)]["NAV_SAMPLE_RATE"] = b"NAV_SAMPLE_RATE               =  20\n"

        _fd, fname = mkstemp()
        fd = open(_fd)

        self.filename = fname
        for _, arr in sections.items():
            arr.tofile(fd)
        fd.close()
        self.fh = eps.EPSAVHRRFile(self.filename, {"start_time": "now",
                                                   "end_time": "later"}, {})

    def test_read_all(self):
        """Test initialization."""
        self.fh._read_all()
        assert self.fh.scanlines == 1080
        assert self.fh.pixels == 2048

    def test_dataset(self):
        """Test getting a dataset."""
        did = make_dataid(name="1", calibration="reflectance")
        res = self.fh.get_dataset(did, {})
        assert isinstance(res, xr.DataArray)
        assert res.attrs["platform_name"] == "Metop-C"
        assert res.attrs["sensor"] == "avhrr-3"
        assert res.attrs["name"] == "1"
        assert res.attrs["calibration"] == "reflectance"
        assert res.attrs["units"] == "%"

        did = make_dataid(name="4", calibration="brightness_temperature")
        res = self.fh.get_dataset(did, {})
        assert isinstance(res, xr.DataArray)
        assert res.attrs["platform_name"] == "Metop-C"
        assert res.attrs["sensor"] == "avhrr-3"
        assert res.attrs["name"] == "4"
        assert res.attrs["calibration"] == "brightness_temperature"
        assert res.attrs["units"] == "K"

    def test_get_dataset_radiance(self):
        """Test loading a data array with radiance calibration."""
        did = make_dataid(name="1", calibration="radiance")
        res = self.fh.get_dataset(did, {})
        assert isinstance(res, xr.DataArray)
        assert res.attrs["platform_name"] == "Metop-C"
        assert res.attrs["sensor"] == "avhrr-3"
        assert res.attrs["name"] == "1"
        assert res.attrs["calibration"] == "radiance"
        assert res.attrs["units"] == "W m^-2 sr^-1"

    def test_navigation(self):
        """Test the navigation."""
        did = make_dataid(name="longitude")
        res = self.fh.get_dataset(did, {})
        assert isinstance(res, xr.DataArray)
        assert res.attrs["platform_name"] == "Metop-C"
        assert res.attrs["sensor"] == "avhrr-3"
        assert res.attrs["name"] == "longitude"

    def test_angles(self):
        """Test the navigation."""
        did = make_dataid(name="solar_zenith_angle")
        res = self.fh.get_dataset(did, {})
        assert isinstance(res, xr.DataArray)
        assert res.attrs["platform_name"] == "Metop-C"
        assert res.attrs["sensor"] == "avhrr-3"
        assert res.attrs["name"] == "solar_zenith_angle"

    def test_clould_flags(self):
        """Test getting the cloud flags."""
        did = make_dataid(name="cloud_flags")
        res = self.fh.get_dataset(did, {})
        assert isinstance(res, xr.DataArray)
        assert res.attrs["platform_name"] == "Metop-C"
        assert res.attrs["sensor"] == "avhrr-3"
        assert res.attrs["name"] == "cloud_flags"

    @mock.patch("satpy.readers.eps_l1b.EPSAVHRRFile.__getitem__")
    def test_get_full_angles_twice(self, mock__getitem__):
        """Test get full angles twice."""
        geotiemock = mock.Mock()
        metop20kmto1km = geotiemock.metop20kmto1km
        metop20kmto1km.side_effect = lambda x, y: (x.copy(), y.copy())

        def mock_getitem(key):
            data = {"ANGULAR_RELATIONS_FIRST": np.zeros((7, 4)),
                    "ANGULAR_RELATIONS": np.zeros((7, 103, 4)),
                    "ANGULAR_RELATIONS_LAST": np.zeros((7, 4)),
                    "NAV_SAMPLE_RATE": 20}
            return data[key]
        mock__getitem__.side_effect = mock_getitem

        avhrr_reader = satpy.readers.eps_l1b.EPSAVHRRFile(
            filename="foo",
            filename_info={"start_time": "foo", "end_time": "bar"},
            filetype_info={"foo": "bar"}
        )
        avhrr_reader.scanlines = 7
        avhrr_reader.pixels = 2048

        with mock.patch.dict("sys.modules", geotiepoints=geotiemock):
            # Get dask arrays
            sun_azi, sun_zen, sat_azi, sat_zen = avhrr_reader.get_full_angles()
            # Convert to numpy array
            sun_zen_np1 = np.array(sun_zen)
            # Convert to numpy array again
            sun_zen_np2 = np.array(sun_zen)
            assert np.allclose(sun_zen_np1, sun_zen_np2)


class TestWrongScanlinesEPSL1B(BaseTestCaseEPSL1B):
    """Test the filehandler on a corrupt file."""

    @pytest.fixture(autouse=True)
    def _inject_fixtures(self, caplog):
        """Inject caplog."""
        self._caplog = caplog

    def setUp(self):
        """Set up the tests."""
        # ipr is not present in the xml format ?
        self.scan_lines = 1080
        self.earth_views = 2048

        sections = self._create_structure()
        sections[("mphr", 0)]["TOTAL_MDR"] = (b"TOTAL_MDR                     =   " +
                                              bytes(str(self.scan_lines - 2), encoding="ascii") +
                                              b"\n")
        sections[("mphr", 0)]["SPACECRAFT_ID"] = b"SPACECRAFT_ID                 = M03\n"
        sections[("mphr", 0)]["INSTRUMENT_ID"] = b"INSTRUMENT_ID                 = AVHR\n"
        sections[("sphr", 0)]["EARTH_VIEWS_PER_SCANLINE"] = (b"EARTH_VIEWS_PER_SCANLINE      =  " +
                                                             bytes(str(self.earth_views), encoding="ascii") +
                                                             b"\n")
        sections[("sphr", 0)]["NAV_SAMPLE_RATE"] = b"NAV_SAMPLE_RATE               =  20\n"
        _fd, fname = mkstemp()
        fd = open(_fd)

        self.filename = fname
        for _, arr in sections.items():
            arr.tofile(fd)
        fd.close()
        self.fh = eps.EPSAVHRRFile(self.filename, {"start_time": "now",
                                                   "end_time": "later"}, {})

    def test_read_all_return_right_number_of_scan_lines(self):
        """Test scanline assignment."""
        self.fh._read_all()
        assert self.fh.scanlines == self.scan_lines

    def test_read_all_warns_about_scan_lines(self):
        """Test scanline assignment."""
        self.fh._read_all()
        assert "scanlines" in self._caplog.records[0].message
        assert self._caplog.records[0].levelname == "WARNING"

    def test_read_all_assigns_int_scan_lines(self):
        """Test scanline assignment."""
        self.fh._read_all()
        assert isinstance(self.fh.scanlines, int)

    def test_get_dataset_longitude_shape_is_right(self):
        """Test that the shape of longitude is 1080."""
        key = make_dataid(name="longitude")
        longitudes = self.fh.get_dataset(key, dict())
        assert longitudes.shape == (self.scan_lines, self.earth_views)

    def tearDown(self):
        """Tear down the tests."""
        with suppress(OSError):
            os.remove(self.filename)


class TestWrongSamplingEPSL1B(BaseTestCaseEPSL1B):
    """Test the filehandler on a corrupt file."""

    @pytest.fixture(autouse=True)
    def _inject_fixtures(self, caplog):
        """Inject caplog."""
        self._caplog = caplog

    def setUp(self):
        """Set up the tests."""
        self.scan_lines = 1080
        self.earth_views = 2048
        self.sample_rate = 23
        sections = self._create_structure()
        sections[("mphr", 0)]["TOTAL_MDR"] = (b"TOTAL_MDR                     =   " +
                                              bytes(str(self.scan_lines), encoding="ascii") +
                                              b"\n")
        sections[("mphr", 0)]["SPACECRAFT_ID"] = b"SPACECRAFT_ID                 = M03\n"
        sections[("mphr", 0)]["INSTRUMENT_ID"] = b"INSTRUMENT_ID                 = AVHR\n"
        sections[("sphr", 0)]["EARTH_VIEWS_PER_SCANLINE"] = (b"EARTH_VIEWS_PER_SCANLINE      =  " +
                                                             bytes(str(self.earth_views), encoding="ascii") +
                                                             b"\n")
        sections[("sphr", 0)]["NAV_SAMPLE_RATE"] = (b"NAV_SAMPLE_RATE               =  " +
                                                    bytes(str(self.sample_rate), encoding="ascii") +
                                                    b"\n")
        _fd, fname = mkstemp()
        fd = open(_fd)

        self.filename = fname
        for _, arr in sections.items():
            arr.tofile(fd)
        fd.close()
        self.fh = eps.EPSAVHRRFile(self.filename, {"start_time": "now",
                                                   "end_time": "later"}, {})

    def test_get_dataset_fails_because_of_wrong_sample_rate(self):
        """Test that lons fail to be interpolate."""
        key = make_dataid(name="longitude")
        with pytest.raises(NotImplementedError):
            self.fh.get_dataset(key, dict())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to TROPOMI L2 Reader.

The TROPOspheric Monitoring Instrument (TROPOMI) is the satellite instrument
on board the Copernicus Sentinel-5 Precursor satellite. It measures key
atmospheric trace gasses, such as ozone, nitrogen oxides, sulfur dioxide,
carbon monoxide, methane, and formaldehyde.

Level 2 data products are available via the Copernicus Open Access Hub.
For more information visit the following URL:
http://www.tropomi.eu/data-products/level-2-products

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.core.netcdf import NetCDF4FileHandler, netCDF4
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)
DATE_FMT = "%Y-%m-%dT%H:%M:%SZ"
CHUNK_SIZE = get_legacy_chunk_size()


class TROPOMIL2FileHandler(NetCDF4FileHandler):
    """File handler for TROPOMI L2 netCDF files."""

    @property
    def start_time(self):
        """Get start time."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get end time."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def platform_shortname(self):
        """Get platform shortname."""
        return self.filename_info["platform_shortname"]

    @property
    def time_coverage_start(self):
        """Get time_coverage_start."""
        return dt.datetime.strptime(self["/attr/time_coverage_start"], DATE_FMT)

    @property
    def time_coverage_end(self):
        """Get time_coverage_end."""
        return dt.datetime.strptime(self["/attr/time_coverage_end"], DATE_FMT)

    @property
    def sensor(self):
        """Get sensor."""
        res = self["/attr/sensor"]
        if isinstance(res, np.ndarray):
            return str(res.astype(str)).lower()
        return res.lower()

    @property
    def sensor_names(self):
        """Get sensor set."""
        return {self.sensor}

    def available_datasets(self, configured_datasets=None):
        """Automatically determine datasets provided by this file."""
        logger.debug("Available_datasets begin...")

        # Determine shape of the geolocation data (lat/lon)
        lat_shape = None
        for var_name, _val in self.file_content.items():
            # Could probably avoid this hardcoding, will think on it
            if (var_name == "PRODUCT/latitude"):
                lat_shape = self[var_name + "/shape"]
                break

        handled_variables = set()

        # update previously configured datasets
        logger.debug("Starting previously configured variables loop...")
        # if bounds exists, we can assemble them later
        bounds_exist = "latitude_bounds" in self and "longitude_bounds" in self
        for is_avail, ds_info in (configured_datasets or []):

            # some other file handler knows how to load this
            if is_avail is not None:
                yield is_avail, ds_info

            var_name = ds_info.get("file_key", ds_info["name"])
            # logger.debug("Evaluating previously configured variable: %s", var_name)
            matches = self.file_type_matches(ds_info["file_type"])
            # we can confidently say that we can provide this dataset and can
            # provide more info
            assembled = var_name in ["assembled_lat_bounds", "assembled_lon_bounds"]
            if (matches and var_name in self) or (assembled and bounds_exist):
                logger.debug("Handling previously configured variable: %s", var_name)
                if not assembled:
                    # Because assembled variables and bounds use the same file_key,
                    # we need to omit file_key once.
                    handled_variables.add(var_name)
                new_info = ds_info.copy()  # don't mess up the above yielded
                yield True, new_info
            elif is_avail is None:
                # if we didn't know how to handle this dataset and no one else did
                # then we should keep it going down the chain
                yield is_avail, ds_info

        yield from self._iterate_over_dataset_contents(handled_variables, lat_shape)

    def _iterate_over_dataset_contents(self, handled_variables, shape):
        """Iterate over dataset contents.

        This is where we dynamically add new datasets
        We will sift through all groups and variables, looking for data matching
        the geolocation bounds
        """
        for var_name, val in self.file_content.items():
            # Only evaluate variables
            if isinstance(val, netCDF4.Variable):
                logger.debug("Evaluating new variable: %s", var_name)
                var_shape = self[var_name + "/shape"]
                logger.debug("Dims:{}".format(var_shape))
                if shape == var_shape[:len(shape)]:
                    logger.debug("Found valid additional dataset: %s", var_name)
                    # Skip anything we have already configured
                    if var_name in handled_variables:
                        logger.debug("Already handled, skipping: %s", var_name)
                        continue
                    handled_variables.add(var_name)
                    last_index_separator = var_name.rindex("/")
                    last_index_separator = last_index_separator + 1
                    var_name_no_path = var_name[last_index_separator:]
                    logger.debug("Using short name of: %s", var_name_no_path)
                    # Create new ds_info object
                    if var_name_no_path in ["latitude_bounds", "longitude_bounds"]:
                        coordinates = []
                    else:
                        coordinates = ["longitude", "latitude"]
                    new_info = {
                        "name": var_name_no_path,
                        "file_key": var_name,
                        "coordinates": coordinates,
                        "file_type": self.filetype_info["file_type"],
                    }
                    yield True, new_info

    def get_metadata(self, data, ds_info):
        """Get metadata."""
        metadata = {}
        metadata.update(data.attrs)
        metadata.update(ds_info)
        metadata.update({
            "platform_shortname": self.platform_shortname,
            "sensor": self.sensor,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "time_coverage_start": self.time_coverage_start,
            "time_coverage_end": self.time_coverage_end,
        })

        return metadata

    def _rename_dims(self, data_arr):
        """Normalize dimension names with the rest of Satpy."""
        dims_dict = {}
        if "ground_pixel" in data_arr.dims:
            dims_dict["ground_pixel"] = "x"
        if "scanline" in data_arr.dims:
            dims_dict["scanline"] = "y"
        return data_arr.rename(dims_dict)

    def prepare_geo(self, bounds_data):
        """Prepare lat/lon bounds for pcolormesh.

        lat/lon bounds are ordered in the following way::

            3----2
            |    |
            0----1

        Extend longitudes and latitudes with one element to support
        "pcolormesh"::

            (X[i+1, j], Y[i+1, j])         (X[i+1, j+1], Y[i+1, j+1])
                                  +--------+
                                  | C[i,j] |
                                  +--------+
                 (X[i, j], Y[i, j])        (X[i, j+1], Y[i, j+1])

        """
        # Create the left array
        left = np.vstack([bounds_data[:, :, 0], bounds_data[-1:, :, 3]])
        # Create the right array
        right = np.vstack([bounds_data[:, -1:, 1], bounds_data[-1:, -1:, 2]])
        # Stack horizontally
        dest = np.hstack([left, right])
        # Convert to DataArray
        dask_dest = da.from_array(dest, chunks=CHUNK_SIZE)
        dest = xr.DataArray(dask_dest,
                            dims=("y_bounds", "x_bounds"),
                            attrs=bounds_data.attrs
                            )
        return dest

    def get_dataset(self, ds_id, ds_info):
        """Get dataset."""
        logger.debug("Getting data for: %s", ds_id["name"])
        file_key = ds_info.get("file_key", ds_id["name"])
        data = self[file_key]
        data.attrs = self.get_metadata(data, ds_info)
        fill_value = data.attrs.get("_FillValue", np.float32(np.nan))
        data = data.squeeze()

        # preserve integer data types if possible
        if np.issubdtype(data.dtype, np.integer):
            new_fill = fill_value
        else:
            new_fill = np.float32(np.nan)
            data.attrs.pop("_FillValue", None)
        good_mask = data != fill_value

        scale_factor = data.attrs.get("scale_factor")
        add_offset = data.attrs.get("add_offset")
        if scale_factor is not None:
            data = data * scale_factor + add_offset

        data = data.where(good_mask, new_fill)
        data = self._rename_dims(data)

        # drop coords whose units are not meters
        drop_list = ["y", "x", "layer", "vertices"]
        coords_exist = [coord for coord in drop_list if coord in data.coords]
        if coords_exist:
            data = data.drop_vars(coords_exist)

        if ds_id["name"] in ["assembled_lat_bounds", "assembled_lon_bounds"]:
            data = self.prepare_geo(data)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The scmi_abi_l1b reader tests package."""

import unittest
from unittest import mock

import numpy as np
import pytest
import xarray as xr


class FakeDataset(object):
    """Fake dataset."""

    def __init__(self, info, attrs, dims=None):
        """Init the dataset."""
        for var_name, var_data in list(info.items()):
            if isinstance(var_data, np.ndarray):
                info[var_name] = xr.DataArray(var_data)
        self.info = info
        self.attrs = attrs
        self.dims = dims or {}

    def __getitem__(self, key):
        """Get item."""
        return self.info.get(key, self.dims.get(key))

    def __contains__(self, key):
        """Check contains."""
        return key in self.info or key in self.dims

    def rename(self, *args, **kwargs):
        """Rename the dataset."""
        return self

    def close(self):
        """Close the dataset."""
        return


class TestSCMIFileHandler(unittest.TestCase):
    """Test the SCMIFileHandler reader."""

    @mock.patch("satpy.readers.scmi.xr")
    def setUp(self, xr_):
        """Set up for test."""
        from satpy.readers.scmi import SCMIFileHandler
        rad_data = (np.arange(10.).reshape((2, 5)) + 1.)
        rad_data = (rad_data + 1.) / 0.5
        rad_data = rad_data.astype(np.int16)
        self.expected_rad = rad_data.astype(np.float64) * 0.5 + -1.
        self.expected_rad[-1, -2] = np.nan
        time = xr.DataArray(0.)
        rad = xr.DataArray(
            rad_data,
            dims=("y", "x"),
            attrs={
                "scale_factor": 0.5,
                "add_offset": -1.,
                "_FillValue": 20,
                "standard_name": "toa_bidirectional_reflectance",
            },
            coords={
                "time": time,
            }
        )
        xr_.open_dataset.return_value = FakeDataset(
            {
                "Sectorized_CMI": rad,
                "nominal_satellite_subpoint_lat": np.array(0.0),
                "nominal_satellite_subpoint_lon": np.array(-89.5),
                "nominal_satellite_height": np.array(35786.02),
            },
            {
                "start_date_time": "2017210120000",
                "satellite_id": "GOES-16",
                "satellite_longitude": -90.,
                "satellite_latitude": 0.,
                "satellite_altitude": 35785831.,
            },
            {"y": 2, "x": 5},
        )

        self.reader = SCMIFileHandler("filename",
                                      {"platform_shortname": "G16"},
                                      {"filetype": "info"})

    def test_basic_attributes(self):
        """Test getting basic file attributes."""
        import datetime as dt

        from satpy.tests.utils import make_dataid
        assert self.reader.start_time == dt.datetime(2017, 7, 29, 12, 0, 0, 0)
        assert self.reader.end_time == dt.datetime(2017, 7, 29, 12, 0, 0, 0)
        assert self.reader.get_shape(make_dataid(name="C05"), {}) == (2, 5)

    def test_data_load(self):
        """Test data loading."""
        from satpy.tests.utils import make_dataid
        res = self.reader.get_dataset(
            make_dataid(name="C05", calibration="reflectance"), {})

        np.testing.assert_allclose(res.data, self.expected_rad, equal_nan=True)
        assert "scale_factor" not in res.attrs
        assert "_FillValue" not in res.attrs
        assert res.attrs["standard_name"] == "toa_bidirectional_reflectance"
        assert "orbital_parameters" in res.attrs
        orb_params = res.attrs["orbital_parameters"]
        assert orb_params["projection_longitude"] == -90.0
        assert orb_params["projection_latitude"] == 0.0
        assert orb_params["projection_altitude"] == 35785831.0


class TestSCMIFileHandlerArea(unittest.TestCase):
    """Test the SCMIFileHandler's area creation."""

    @mock.patch("satpy.readers.scmi.xr")
    def create_reader(self, proj_name, proj_attrs, xr_):
        """Create a fake reader."""
        from satpy.readers.scmi import SCMIFileHandler
        proj = xr.DataArray([], attrs=proj_attrs)
        x__ = xr.DataArray(
            [0, 1],
            attrs={"scale_factor": 2., "add_offset": -1., "units": "meters"},
        )
        y__ = xr.DataArray(
            [0, 1],
            attrs={"scale_factor": -2., "add_offset": 1., "units": "meters"},
        )
        xr_.open_dataset.return_value = FakeDataset({
            "goes_imager_projection": proj,
            "x": x__,
            "y": y__,
            "Sectorized_CMI": np.ones((2, 2))},
            {
                "satellite_id": "GOES-16",
                "grid_mapping": proj_name,
            },
            {
                "y": y__.size,
                "x": x__.size,
            }
        )

        return SCMIFileHandler("filename",
                               {"platform_shortname": "G16"},
                               {"filetype": "info"})

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_geos(self, adef):
        """Test the area generation for geos projection."""
        reader = self.create_reader(
            "goes_imager_projection",
            {
                "semi_major_axis": 1.,
                "semi_minor_axis": 1.,
                "perspective_point_height": 1.,
                "longitude_of_projection_origin": -90.,
                "sweep_angle_axis": u"x",
                "grid_mapping_name": "geostationary",
            }
        )
        reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"a": 1.0, "b": 1.0, "h": 1.0,
                                "lon_0": -90.0, "lat_0": 0.0,
                                "proj": "geos", "sweep": "x", "units": "m"}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-2., -2., 2, 2.))

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_lcc(self, adef):
        """Test the area generation for lcc projection."""
        reader = self.create_reader(
            "goes_imager_projection",
            {
                "semi_major_axis": 1.,
                "semi_minor_axis": 1.,
                "longitude_of_central_meridian": -90.,
                "standard_parallel": 25.,
                "latitude_of_projection_origin": 25.,
                "grid_mapping_name": "lambert_conformal_conic",
            }
        )
        reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"a": 1.0, "b": 1.0,
                                "lon_0": -90.0, "lat_0": 25.0, "lat_1": 25.0,
                                "proj": "lcc", "units": "m"}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-2., -2., 2, 2.))

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_stere(self, adef):
        """Test the area generation for stere projection."""
        reader = self.create_reader(
            "goes_imager_projection",
            {
                "semi_major_axis": 1.,
                "semi_minor_axis": 1.,
                "straight_vertical_longitude_from_pole": -90.,
                "standard_parallel": 60.,
                "latitude_of_projection_origin": 90.,
                "grid_mapping_name": "polar_stereographic",
            }
        )
        reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"a": 1.0, "b": 1.0,
                                "lon_0": -90.0, "lat_0": 90.0, "lat_ts": 60.0,
                                "proj": "stere", "units": "m"}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-2., -2., 2, 2.))

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_merc(self, adef):
        """Test the area generation for merc projection."""
        reader = self.create_reader(
            "goes_imager_projection",
            {
                "semi_major_axis": 1.,
                "semi_minor_axis": 1.,
                "longitude_of_projection_origin": -90.,
                "standard_parallel": 0.,
                "grid_mapping_name": "mercator",
            }
        )
        reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        assert call_args[3] == {"a": 1.0, "b": 1.0,
                                "lon_0": -90.0, "lat_0": 0.0, "lat_ts": 0.0,
                                "proj": "merc", "units": "m"}
        assert call_args[4] == reader.ncols
        assert call_args[5] == reader.nlines
        np.testing.assert_allclose(call_args[6], (-2., -2., 2, 2.))

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def_bad(self, adef):
        """Test the area generation for bad projection."""
        reader = self.create_reader(
            "goes_imager_projection",
            {
                "semi_major_axis": 1.,
                "semi_minor_axis": 1.,
                "longitude_of_projection_origin": -90.,
                "standard_parallel": 0.,
                "grid_mapping_name": "fake",
            }
        )
        with pytest.raises(ValueError, match="Can't handle projection 'fake'"):
            reader.get_area_def(None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Modifiers related to atmospheric corrections or adjustments."""

import logging

import dask.array as da
import numpy as np
import xarray as xr

from satpy.modifiers import ModifierBase
from satpy.modifiers._crefl import ReflectanceCorrector  # noqa
from satpy.modifiers.angles import compute_relative_azimuth, get_angles, get_satellite_zenith_angle

logger = logging.getLogger(__name__)


class PSPRayleighReflectance(ModifierBase):
    """Pyspectral-based rayleigh corrector for visible channels.

    It is possible to use ``reduce_lim_low``, ``reduce_lim_high`` and ``reduce_strength``
    together to reduce rayleigh correction at high solar zenith angle and make the image
    transition from rayleigh-corrected to partially/none rayleigh-corrected at day/night edge,
    therefore producing a more natural look, which could be especially helpful for geostationary
    satellites. This reduction starts at solar zenith angle of ``reduce_lim_low``, and ends in
    ``reduce_lim_high``. It's linearly scaled between these two angles. The ``reduce_strength``
    controls the amount of the reduction. When the solar zenith angle reaches ``reduce_lim_high``,
    the rayleigh correction will remain ``(1 - reduce_strength)`` of its initial reduce_strength
    at ``reduce_lim_high``.

    To use this function in a YAML configuration file:

    .. code-block:: yaml

      rayleigh_corrected_reduced:
        modifier: !!python/name:satpy.modifiers.PSPRayleighReflectance
        atmosphere: us-standard
        aerosol_type: rayleigh_only
        reduce_lim_low: 70
        reduce_lim_high: 95
        reduce_strength: 0.6
        prerequisites:
          - name: B03
            modifiers: [sunz_corrected]
        optional_prerequisites:
          - satellite_azimuth_angle
          - satellite_zenith_angle
          - solar_azimuth_angle
          - solar_zenith_angle

    In the case above, rayleigh correction is reduced gradually starting at solar zenith angle 70°.
    When reaching 95°, the correction will only remain 40% its initial strength at 95°.

    """

    def __call__(self, projectables, optional_datasets=None, **info):
        """Get the corrected reflectance when removing Rayleigh scattering.

        Uses pyspectral.
        """
        from pyspectral.rayleigh import Rayleigh
        projectables = projectables + (optional_datasets or [])
        if len(projectables) != 6:
            vis, red = self.match_data_arrays(projectables)
            # Adjust the angle data precision to match the data
            # This does not affect the accuracy visibly
            sata, satz, suna, sunz = [d.astype(vis.dtype) for d in get_angles(vis)]
        else:
            vis, red, sata, satz, suna, sunz = self.match_data_arrays(projectables)
            # First make sure the two azimuth angles are in the range 0-360:
            sata = sata % 360.
            suna = suna % 360.

        # get the dask array underneath
        sata = sata.data
        satz = satz.data
        suna = suna.data
        sunz = sunz.data

        ssadiff = compute_relative_azimuth(sata, suna)
        del sata, suna

        atmosphere = self.attrs.get("atmosphere", "us-standard")
        aerosol_type = self.attrs.get("aerosol_type", "marine_clean_aerosol")
        reduce_lim_low = abs(self.attrs.get("reduce_lim_low", 70))
        reduce_lim_high = abs(self.attrs.get("reduce_lim_high", 105))
        reduce_strength = np.clip(self.attrs.get("reduce_strength", 0), 0, 1).astype(vis.dtype)

        logger.info("Removing Rayleigh scattering with atmosphere '%s' and "
                    "aerosol type '%s' for '%s'",
                    atmosphere, aerosol_type, vis.attrs["name"])
        corrector = Rayleigh(vis.attrs["platform_name"], vis.attrs["sensor"],
                             atmosphere=atmosphere,
                             aerosol_type=aerosol_type)

        try:
            refl_cor_band = corrector.get_reflectance(sunz, satz, ssadiff,
                                                      vis.attrs["name"],
                                                      red.data)
        except (KeyError, IOError):
            logger.warning("Could not get the reflectance correction using band name: %s", vis.attrs["name"])
            logger.warning("Will try use the wavelength, however, this may be ambiguous!")
            refl_cor_band = corrector.get_reflectance(sunz, satz, ssadiff,
                                                      vis.attrs["wavelength"][1],
                                                      red.data)

        if reduce_strength > 0:
            if reduce_lim_low > reduce_lim_high:
                reduce_lim_low = reduce_lim_high
            refl_cor_band = corrector.reduce_rayleigh_highzenith(sunz, refl_cor_band,
                                                                 reduce_lim_low, reduce_lim_high, reduce_strength)

        proj = vis - refl_cor_band
        proj.attrs = vis.attrs
        self.apply_modifier_info(vis, proj)
        return proj


def _call_mapped_correction(satz, band_data, corrector, band_name):
    # need to convert to masked array
    orig_dtype = band_data.dtype
    band_data = np.ma.masked_where(np.isnan(band_data), band_data)
    res = corrector.get_correction(satz, band_name, band_data)
    return res.filled(np.nan).astype(orig_dtype, copy=False)


class PSPAtmosphericalCorrection(ModifierBase):
    """Correct for atmospheric effects."""

    def __call__(self, projectables, optional_datasets=None, **info):
        """Get the atmospherical correction.

        Uses pyspectral.
        """
        from pyspectral.atm_correction_ir import AtmosphericalCorrection

        band = projectables[0]

        if optional_datasets:
            satz = optional_datasets[0]
        else:
            satz = get_satellite_zenith_angle(band)
        satz = satz.data  # get dask array underneath

        logger.info("Correction for limb cooling")
        corrector = AtmosphericalCorrection(band.attrs["platform_name"],
                                            band.attrs["sensor"])

        atm_corr = da.map_blocks(_call_mapped_correction, satz, band.data,
                                 corrector=corrector,
                                 band_name=band.attrs["name"],
                                 meta=np.array((), dtype=band.dtype))
        proj = xr.DataArray(atm_corr, attrs=band.attrs,
                            dims=band.dims, coords=band.coords)
        self.apply_modifier_info(band, proj)

        return proj


class CO2Corrector(ModifierBase):
    """CO2 correction of the brightness temperature of the MSG 3.9um channel.

    .. math::

      T4_CO2corr = (BT(IR3.9)^4 + Rcorr)^0.25
      Rcorr = BT(IR10.8)^4 - (BT(IR10.8)-dt_CO2)^4
      dt_CO2 = (BT(IR10.8)-BT(IR13.4))/4.0

    Derived from D. Rosenfeld, "CO2 Correction of Brightness Temperature of Channel IR3.9"
    References:
        - https://resources.eumetrain.org/IntGuide/PowerPoints/Channels/conversion.ppt
    """

    def __call__(self, projectables, optional_datasets=None, **info):
        """Apply correction."""
        ir_039, ir_108, ir_134 = projectables
        logger.info("Applying CO2 correction")
        dt_co2 = (ir_108 - ir_134) / 4.0
        rcorr = ir_108 ** 4 - (ir_108 - dt_co2) ** 4
        t4_co2corr = (ir_039 ** 4 + rcorr).clip(0.0) ** 0.25

        t4_co2corr.attrs = ir_039.attrs.copy()

        self.apply_modifier_info(ir_039, t4_co2corr)

        return t4_co2corr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2015-2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Composite classes for the ABI instrument."""

import logging

from satpy.composites.core import GenericCompositor

LOG = logging.getLogger(__name__)


class SimulatedGreen(GenericCompositor):
    """A single-band dataset resembling a Green (0.55 µm) band.

    This compositor creates a single band product by combining three
    other bands in various amounts. The general formula with
    dependencies (d) and fractions (f) is::

        result = d1 * f1 + d2 * f2 + d3 * f3

    See the `fractions` keyword argument for more information.
    Common used fractions for ABI data with C01, C02, and C03 inputs include:

    - SatPy default (historical): (0.465, 0.465, 0.07)
    - `CIMSS (Kaba) <https://agupubs.onlinelibrary.wiley.com/doi/10.1029/2018EA000379>`_: (0.45, 0.45, 0.10)
    - `EDC <http://edc.occ-data.org/goes16/python/>`_: (0.45706946, 0.48358168, 0.06038137)

    """

    def __init__(self, name, fractions=(0.465, 0.465, 0.07), **kwargs):  # noqa: D417
        """Initialize fractions for input channels.

        Args:
            name (str): Name of this composite
            fractions (Iterable): Fractions of each input band to include in the result.

        """
        self.fractions = fractions
        super(SimulatedGreen, self).__init__(name, **kwargs)

    def __call__(self, projectables, optional_datasets=None, **attrs):
        """Generate the single band composite."""
        c01, c02, c03 = self.match_data_arrays(projectables)
        res = c01 * self.fractions[0] + c02 * self.fractions[1] + c03 * self.fractions[2]
        res.attrs = c03.attrs.copy()
        return super(SimulatedGreen, self).__call__((res,), **attrs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.geocat module."""

import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler
from satpy.tests.utils import convert_file_content_to_data_array

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)


class FakeNetCDF4FileHandler2(FakeNetCDF4FileHandler):
    """Swap-in NetCDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {
            "/attr/Platform_Name": filename_info["platform_shortname"],
            "/attr/Element_Resolution": 2.,
            "/attr/Line_Resolution": 2.,
            "/attr/Subsatellite_Longitude": -70.2 if "GOES" in filename_info["platform_shortname"] else 140.65,
            "pixel_longitude": DEFAULT_LON_DATA,
            "pixel_longitude/attr/scale_factor": 1.,
            "pixel_longitude/attr/add_offset": 0.,
            "pixel_longitude/shape": DEFAULT_FILE_SHAPE,
            "pixel_longitude/attr/_FillValue": np.nan,
            "pixel_latitude": DEFAULT_LAT_DATA,
            "pixel_latitude/attr/scale_factor": 1.,
            "pixel_latitude/attr/add_offset": 0.,
            "pixel_latitude/shape": DEFAULT_FILE_SHAPE,
            "pixel_latitude/attr/_FillValue": np.nan,
        }
        sensor = {
            "HIMAWARI-8": "himawari8",
            "GOES-17": "goesr",
            "GOES-16": "goesr",
            "GOES-13": "goes",
            "GOES-14": "goes",
            "GOES-15": "goes",
        }[filename_info["platform_shortname"]]
        file_content["/attr/Sensor_Name"] = sensor

        if filename_info["platform_shortname"] == "HIMAWARI-8":
            file_content["pixel_longitude"] = DEFAULT_LON_DATA + 130.

        file_content["variable1"] = DEFAULT_FILE_DATA.astype(np.float32)
        file_content["variable1/attr/_FillValue"] = -1
        file_content["variable1/attr/scale_factor"] = 1.
        file_content["variable1/attr/add_offset"] = 0.
        file_content["variable1/attr/units"] = "1"
        file_content["variable1/shape"] = DEFAULT_FILE_SHAPE

        # data with fill values
        file_content["variable2"] = np.ma.masked_array(
            DEFAULT_FILE_DATA.astype(np.float32),
            mask=np.zeros_like(DEFAULT_FILE_DATA))
        file_content["variable2"].mask[::5, ::5] = True
        file_content["variable2/attr/_FillValue"] = -1
        file_content["variable2/attr/scale_factor"] = 1.
        file_content["variable2/attr/add_offset"] = 0.
        file_content["variable2/attr/units"] = "1"
        file_content["variable2/shape"] = DEFAULT_FILE_SHAPE

        # category
        file_content["variable3"] = DEFAULT_FILE_DATA.astype(np.byte)
        file_content["variable3/attr/_FillValue"] = -128
        file_content["variable3/attr/flag_meanings"] = "clear water supercooled mixed ice unknown"
        file_content["variable3/attr/flag_values"] = [0, 1, 2, 3, 4, 5]
        file_content["variable3/attr/units"] = "1"
        file_content["variable3/shape"] = DEFAULT_FILE_SHAPE

        attrs = ("_FillValue", "flag_meanings", "flag_values", "units")
        convert_file_content_to_data_array(
            file_content, attrs=attrs,
            dims=("z", "lines", "elements"))
        return file_content


class TestGEOCATReader(unittest.TestCase):
    """Test GEOCAT Reader."""

    yaml_file = "geocat.yaml"

    def setUp(self):
        """Wrap NetCDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.geocat import GEOCATFileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(GEOCATFileHandler, "__bases__", (FakeNetCDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "geocatL2.GOES-13.2015143.234500.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_init_with_kwargs(self):
        """Test basic init with extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs, xarray_kwargs={"decode_times": True})
        loadables = r.select_files_from_pathnames([
            "geocatL2.GOES-13.2015143.234500.nc",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables, fh_kwargs={"xarray_kwargs": {"decode_times": True}})
        # make sure we have some files
        assert r.file_handlers

    def test_load_all_old_goes(self):
        """Test loading all test datasets from old GOES files."""
        import xarray as xr

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.geocat.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "geocatL2.GOES-13.2015143.234500.nc",
            ])
            r.create_filehandlers(loadables)
        datasets = r.load(["variable1",
                           "variable2",
                           "variable3"])
        assert len(datasets) == 3
        for v in datasets.values():
            assert "calibration" not in v.attrs
            assert v.attrs["units"] == "1"
        assert datasets["variable3"].attrs.get("flag_meanings") is not None

    def test_load_all_himawari8(self):
        """Test loading all test datasets from H8 NetCDF file."""
        import xarray as xr
        from pyresample.geometry import AreaDefinition

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.geocat.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "geocatL2.HIMAWARI-8.2017092.210730.R304.R20.nc",
            ])
            r.create_filehandlers(loadables)
        datasets = r.load(["variable1",
                           "variable2",
                           "variable3"])
        assert len(datasets) == 3
        for v in datasets.values():
            assert "calibration" not in v.attrs
            assert v.attrs["units"] == "1"
        assert datasets["variable3"].attrs.get("flag_meanings") is not None
        assert isinstance(datasets["variable1"].attrs["area"], AreaDefinition)

    def test_load_all_goes17_hdf4(self):
        """Test loading all test datasets from GOES-17 HDF4 file."""
        import xarray as xr
        from pyresample.geometry import AreaDefinition

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.geocat.netCDF4.Variable", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "geocatL2.GOES-17.CONUS.2020041.163130.hdf",
            ])
            r.create_filehandlers(loadables)
        datasets = r.load(["variable1",
                           "variable2",
                           "variable3"])
        assert len(datasets) == 3
        for v in datasets.values():
            assert "calibration" not in v.attrs
            assert v.attrs["units"] == "1"
        assert datasets["variable3"].attrs.get("flag_meanings") is not None
        assert isinstance(datasets["variable1"].attrs["area"], AreaDefinition)



#!/usr/bin/python
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unittests for Landsat image readers."""

import os
import shutil
from datetime import datetime, timezone

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition
from pytest_lazy_fixtures.lazy_fixture import lf

from satpy import Scene
from satpy.readers.core.landsat import (
    ETMCHReader,
    ETML2CHReader,
    LandsatL1MDReader,
    LandsatL2MDReader,
    MSSCHReader,
    OLITIRSCHReader,
    OLITIRSL2CHReader,
    TMCHReader,
    TML2CHReader,
)

ETC_DIR = os.path.join(os.path.dirname(__file__), "landsat_metadata")

x_size = 100
y_size = 100

oli_tirs_l1_date = datetime(2024, 5, 12, tzinfo=timezone.utc)
oli_tirs_l2_date = datetime(2024, 6, 16, tzinfo=timezone.utc)
etm_l1_date = datetime(2023, 12, 8, tzinfo=timezone.utc)
etm_l2_date = datetime(2023, 8, 17, tzinfo=timezone.utc)
tm_l1_date = datetime(1989, 8, 18, tzinfo=timezone.utc)
tm_l2_date = datetime(2011, 8, 17, tzinfo=timezone.utc)
mss_l1_landsat1_date = datetime(1972, 7, 29, tzinfo=timezone.utc)
mss_l1_landsat4_date = datetime(1984, 4, 15, tzinfo=timezone.utc)

oli_tirs_l1_datetime = datetime(2024, 5, 2, 18, 0, 24, tzinfo=timezone.utc)
oli_tirs_l2_datetime = datetime(2024, 6, 16, 17, 10, 58, tzinfo=timezone.utc)
etm_l1_datetime = datetime(2023, 12, 8, 11, 44, 51, tzinfo=timezone.utc)
etm_l2_datetime = datetime(2023, 8, 17, 14, 54, 20, tzinfo=timezone.utc)
tm_l1_datetime = datetime(1989, 8, 18, 4, 26, 11, tzinfo=timezone.utc)
tm_l2_datetime = datetime(2011, 8, 17, 7, 10, 40, tzinfo=timezone.utc)
mss_l1_landsat1_datetime = datetime(1972, 7, 29, 16, 49, 31, tzinfo=timezone.utc)
mss_l1_landsat4_datetime = datetime(1984, 4, 15, 16, 38, 15, tzinfo=timezone.utc)

oli_tirs_l1_extent = (619485.0, 2440485.0, 850515.0, 2675715.0)
oli_tirs_l2_extent = (534885.0, 4665585.0, 765015.0, 4899315.0)
etm_l1_extent = (240885.0, -3301215.0, 483015.0, -3091485.0)
etm_l2_extent = (85785.0, 4680885.0, 336915.0, 4904115.0)
tm_l1_extent = (322185.0, 6085185.0, 567615.0, 6311415.0)
tm_l2_extent = (258285.0, 855285.0, 495315.0, 1064415.0)
mss_l1_landsat1_extent = (442110.0, 4667550.0, 673530.0, 4887990.0)
mss_l1_landsat4_extent = (540030.0, 4679970.0, 774450.0, 4888350.0)
antarctic_extent = (-138315.0, 590385.0, 61215.0, 808815.0)

oli_tirs_l1_pan_extent = (619492.5, 2440492.5, 850507.5, 2675707.5)
etm_l1_pan_extent = (240892.5, -3301207.5, 483007.5, -3091492.5)
antarctic_pan_extent = (-138307.5, 590392.5, 61207.5, 808807.5)

oli_tirs_l1_cal_dict = {"B1": (0.012357, -61.78647, 2e-05, -0.1),
                        "B5": (0.0060172, -30.08607, 2e-05, -0.1),
                        "B10": (0.0003342, 0.1, 774.8853, 1321.0789)}
oli_tirs_l2_cal_dict = {"B1": (2.75e-05, -0.2),
                        "B5": (2.75e-05, -0.2),
                        "B10": (0.00341802, 149.0)}
etm_l1_cal_dict = {"B1": (7.7874e-01, -6.97874, 1.1661e-03, -0.010450),
                   "B5": (1.2622e-01, -1.12622, 1.7365e-03, -0.015494),
                   "B6_VCID_2": (3.7205e-02, 3.16280, 666.09, 1282.71)}
etm_l2_cal_dict = {"B1": (2.75e-05, -0.2),
                   "B5": (2.75e-05, -0.2),
                   "B6": (0.00341802, 149.0)}
tm_l1_cal_dict = {"B1": (6.7921e-01, -2.19921, 1.1252e-03, -0.003643),
                  "B5": (1.2508e-01, -0.49508, 1.8160e-03, -0.007188),
                  "B6": (5.5375e-02, 1.18243, 671.62, 1284.30)}
tm_l2_cal_dict = {"B1": (2.75e-05, -0.2),
                  "B5": (2.75e-05, -0.2),
                  "B6": (0.00341802, 149.0)}
mss_l1_landsat1_cal_dict = {"B5": (6.4843e-01, -0.74843, 1.3660e-03, -0.001577),
                            "B6": (6.5236e-01, -0.75236, 1.6580e-03, -0.001912),
                            "B7": (6.0866e-01, -0.60866, 2.3287e-03, -0.002329)}
mss_l1_landsat4_cal_dict = {"B1": (8.7520e-01, 2.92480, 1.5680e-03, 0.005240),
                            "B2": (6.2008e-01, 3.07992, 1.2865e-03, 0.006390),
                            "B3": (5.4921e-01, 4.55079, 1.4070e-03, 0.011659)}

oli_tirs_l1_sp_params = [0.0098329, -49.16426, 2e-05, -0.1]
oli_tirs_l1_th_params = [0.0003342, 0.100000, 480.8883, 1201.1442]
oli_tirs_l2_sp_params = [2.75e-05, -0.2]
oli_tirs_l2_th_params = [0.00341802, 149.0]
etm_l1_sp_params = [9.6929e-01, -6.06929, 2.7591e-03, -0.017277]
etm_l1_th_params = [6.7087e-02, -0.06709, 666.09, 1282.71]
etm_l2_sp_params = [2.75e-05, -0.2]
etm_l2_th_params = [0.00341802, 149.0]
tm_l1_sp_params = [8.7602e-01, -2.38602, 2.7296e-03, -0.007435]
tm_l1_th_params = [5.5375e-02, 1.18243, 671.62, 1284.30]
tm_l2_sp_params = [2.75e-05, -0.2]
tm_l2_th_params = [0.00341802, 149.0]
mss_l1_landsat1_sp_params = [9.5591e-01, -18.55591, 1.7282e-03, -0.033547]
mss_l1_landsat4_sp_params = [4.7638e-01, 3.82362, 1.7954e-03, 0.014411]


@pytest.fixture(scope="module")
def spectral_data():
    """Get the data for the generic spectral channel."""
    return da.random.randint(12000, 16000,
                             size=(y_size, x_size),
                             chunks=(50, 50)).astype(np.uint16)


@pytest.fixture(scope="module")
def thermal_data():
    """Get the data for the generic thermal channel."""
    return da.random.randint(8000, 14000,
                             size=(y_size, x_size),
                             chunks=(50, 50)).astype(np.uint16)


@pytest.fixture(scope="module")
def sza_rad_data():
    """Get the data for the sza or radiance channel."""
    return da.random.randint(1, 10000,
                             size=(y_size, x_size),
                             chunks=(50, 50)).astype(np.uint16)


def create_tif_file(data, name, area, filename, date):
    """Create a tif file."""
    data_array = xr.DataArray(data,
                              dims=("y", "x"),
                              attrs={"name": name,
                                     "start_time": date})
    scn = Scene()
    scn["band_data"] = data_array
    scn["band_data"].attrs["area"] = area
    scn.save_dataset("band_data", writer="geotiff", enhance=False, fill_value=0,
                     filename=os.fspath(filename))


def get_area_fixture(name, zone, extent):
    """Get area def."""
    @pytest.fixture(scope="module")
    def _fixture():
        """Fixture body."""
        pcs_id = "WGS84 / UTM zone {zone}N"
        proj4_dict = {"proj": "utm", "zone": zone, "datum": "WGS84", "units": "m", "no_defs": None, "type": "crs"}
        area_extent = extent
        _fixture.__name__ = f"{name}_area"
        return AreaDefinition("geotiff_area", pcs_id, pcs_id, proj4_dict, x_size, y_size, area_extent)
    _fixture.__name__ = f"{name}_area"
    return _fixture


def make_temp_path_fixture(name):
    """Create a temp directory for a Landsat product."""
    @pytest.fixture(scope="module")
    def _fixture(tmp_path_factory):
        return tmp_path_factory.mktemp(f"{name}_files")
    _fixture.__name__ = f"{name}_files_path"
    return _fixture


def convert_to_fsfile(files):
    """Turn pathes to FSFile objects."""
    from fsspec.implementations.local import LocalFileSystem

    from satpy.readers.core.remote import FSFile

    fs = LocalFileSystem()
    files = (
        FSFile(os.path.abspath(file), fs=fs)
        for file in files
    )
    return files


oli_tirs_l1_area = get_area_fixture("oli_tirs_l1", 40, oli_tirs_l1_extent)
oli_tirs_l2_area = get_area_fixture("oli_tirs_l2", 14, oli_tirs_l2_extent)
etm_l1_area = get_area_fixture("etm_l1", 20, etm_l1_extent)
etm_l2_area = get_area_fixture("etm_l2", 15, etm_l2_extent)
tm_l1_area = get_area_fixture("tm_l1", 46, tm_l1_extent)
tm_l2_area = get_area_fixture("tm_l2", 38, tm_l2_extent)
mss_l1_landsat1_area = get_area_fixture("mss_l1_landsat1", 14, mss_l1_landsat1_extent)
mss_l1_landsat4_area = get_area_fixture("mss_l1_landsat4", 14, mss_l1_landsat4_extent)

oli_tirs_l1_files_path = make_temp_path_fixture("oli_tirs_l1")
oli_tirs_l2_files_path = make_temp_path_fixture("oli_tirs_l2")
etm_l1_files_path = make_temp_path_fixture("etm_l1")
etm_l2_files_path = make_temp_path_fixture("etm_l2")
tm_l1_files_path = make_temp_path_fixture("tm_l1")
tm_l2_files_path = make_temp_path_fixture("tm_l2")
mss_l1_landsat1_files_path = make_temp_path_fixture("mss_l1_landsat1")
mss_l1_landsat4_files_path = make_temp_path_fixture("mss_l1_landsat4")
antarctic_files_path = make_temp_path_fixture("antarctic")


@pytest.fixture(scope="module")
def oli_tirs_l1_b4_file(oli_tirs_l1_files_path, spectral_data, oli_tirs_l1_area):
    """Create the file for the Landsat OLI TIRS L1 B4 channel."""
    filename = oli_tirs_l1_files_path / "LC08_L1TP_026200_20240502_20240513_02_T2_B4.TIF"
    create_tif_file(spectral_data, "B4", oli_tirs_l1_area, filename, oli_tirs_l1_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def oli_tirs_l1_b11_file(oli_tirs_l1_files_path, thermal_data, oli_tirs_l1_area):
    """Create the file for the Landsat OLI TIRS L1 B11 channel."""
    filename = oli_tirs_l1_files_path / "LC08_L1TP_026200_20240502_20240513_02_T2_B11.TIF"
    create_tif_file(thermal_data, "B11", oli_tirs_l1_area, filename, oli_tirs_l1_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def oli_tirs_l1_sza_file(oli_tirs_l1_files_path, sza_rad_data, oli_tirs_l1_area):
    """Create the file for the Landsat OLI TIRS L1 sza."""
    filename = oli_tirs_l1_files_path / "LC08_L1TP_026200_20240502_20240513_02_T2_SZA.TIF"
    create_tif_file(sza_rad_data, "sza", oli_tirs_l1_area, filename, oli_tirs_l1_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def oli_tirs_l1_mda_file(oli_tirs_l1_files_path):
    """Create the Landsat OLI TIRS L1 metadata xml file."""
    filename = oli_tirs_l1_files_path / "LC08_L1TP_026200_20240502_20240513_02_T2_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "oli_tirs_l1_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def oli_tirs_l1_all_files(oli_tirs_l1_b4_file, oli_tirs_l1_b11_file, oli_tirs_l1_mda_file, oli_tirs_l1_sza_file):
    """Return all the files."""
    return oli_tirs_l1_b4_file, oli_tirs_l1_b11_file, oli_tirs_l1_mda_file, oli_tirs_l1_sza_file


@pytest.fixture(scope="module")
def oli_tirs_l2_b4_file(oli_tirs_l2_files_path, spectral_data, oli_tirs_l2_area):
    """Create the file for the Landsat OLI TIRS L2 B4 channel."""
    filename = oli_tirs_l2_files_path / "LC09_L2SP_029030_20240616_20240617_02_T1_SR_B4.TIF"
    create_tif_file(spectral_data, "B4", oli_tirs_l2_area, filename, oli_tirs_l2_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def oli_tirs_l2_b10_file(oli_tirs_l2_files_path, thermal_data, oli_tirs_l2_area):
    """Create the file for the Landsat OLI TIRS L2 B10 channel."""
    filename = oli_tirs_l2_files_path / "LC09_L2SP_029030_20240616_20240617_02_T1_ST_B10.TIF"
    create_tif_file(thermal_data, "B10", oli_tirs_l2_area, filename, oli_tirs_l2_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def oli_tirs_l2_rad_file(oli_tirs_l2_files_path, sza_rad_data, oli_tirs_l2_area):
    """Create the file for the Landsat OLI TIRS L2 TRAD channel."""
    filename = oli_tirs_l2_files_path / "LC09_L2SP_029030_20240616_20240617_02_T1_ST_TRAD.TIF"
    create_tif_file(sza_rad_data, "TRAD", oli_tirs_l2_area, filename, oli_tirs_l2_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def oli_tirs_l2_mda_file(oli_tirs_l2_files_path):
    """Create the Landsat OLI TIRS L2 metadata xml file."""
    filename = oli_tirs_l2_files_path / "LC09_L2SP_029030_20240616_20240617_02_T1_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "oli_tirs_l2_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def oli_tirs_l2_all_files(oli_tirs_l2_b4_file, oli_tirs_l2_b10_file, oli_tirs_l2_mda_file, oli_tirs_l2_rad_file):
    """Return all the files."""
    return oli_tirs_l2_b4_file, oli_tirs_l2_b10_file, oli_tirs_l2_mda_file, oli_tirs_l2_rad_file


@pytest.fixture(scope="module")
def etm_l1_b4_file(etm_l1_files_path, spectral_data, etm_l1_area):
    """Create the file for the Landsat ETM+ L1 B4 channel."""
    filename = etm_l1_files_path / "LE07_L1TP_230080_20231208_20240103_02_T1_B4.TIF"
    create_tif_file(spectral_data, "B4", etm_l1_area, filename, etm_l1_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def etm_l1_b6_file(etm_l1_files_path, thermal_data, etm_l1_area):
    """Create the file for the Landsat ETM+ L1 B6 channel."""
    filename = etm_l1_files_path / "LE07_L1TP_230080_20231208_20240103_02_T1_B6_VCID_1.TIF"
    create_tif_file(thermal_data, "B6_VCID_1", etm_l1_area, filename, etm_l1_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def etm_l1_sza_file(etm_l1_files_path, sza_rad_data, etm_l1_area):
    """Create the file for the Landsat ETM+ L1 sza."""
    filename = etm_l1_files_path / "LE07_L1TP_230080_20231208_20240103_02_T1_SZA.TIF"
    create_tif_file(sza_rad_data, "sza", etm_l1_area, filename, etm_l1_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def etm_l1_mda_file(etm_l1_files_path):
    """Create the Landsat ETM+ L1 metadata xml file."""
    filename = etm_l1_files_path / "LE07_L1TP_230080_20231208_20240103_02_T1_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "etm_l1_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def etm_l1_all_files(etm_l1_b4_file, etm_l1_b6_file, etm_l1_mda_file, etm_l1_sza_file):
    """Return all the files."""
    return etm_l1_b4_file, etm_l1_b6_file, etm_l1_mda_file, etm_l1_sza_file


@pytest.fixture(scope="module")
def etm_l2_b4_file(etm_l2_files_path, spectral_data, etm_l2_area):
    """Create the file for the Landsat ETM+ L2 B4 channel."""
    filename = etm_l2_files_path / "LE07_L2SP_028030_20230817_20230912_02_T1_SR_B4.TIF"
    create_tif_file(spectral_data, "B4", etm_l2_area, filename, etm_l2_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def etm_l2_b6_file(etm_l2_files_path, thermal_data, etm_l2_area):
    """Create the file for the Landsat ETM+ L2 B6 channel."""
    filename = etm_l2_files_path / "LE07_L2SP_028030_20230817_20230912_02_T1_ST_B6.TIF"
    create_tif_file(thermal_data, "B6", etm_l2_area, filename, etm_l2_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def etm_l2_rad_file(etm_l2_files_path, sza_rad_data, etm_l2_area):
    """Create the file for the Landsat ETM+ L2 TRAD channel."""
    filename = etm_l2_files_path / "LE07_L2SP_028030_20230817_20230912_02_T1_ST_TRAD.TIF"
    create_tif_file(sza_rad_data, "TRAD", etm_l2_area, filename, etm_l2_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def etm_l2_mda_file(etm_l2_files_path):
    """Create the Landsat ETM+ L2 metadata xml file."""
    filename = etm_l2_files_path / "LE07_L2SP_028030_20230817_20230912_02_T1_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "etm_l2_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def etm_l2_all_files(etm_l2_b4_file, etm_l2_b6_file, etm_l2_mda_file, etm_l2_rad_file):
    """Return all the files."""
    return etm_l2_b4_file, etm_l2_b6_file, etm_l2_mda_file, etm_l2_rad_file


@pytest.fixture(scope="module")
def tm_l1_b4_file(tm_l1_files_path, spectral_data, tm_l1_area):
    """Create the file for the Landsat TM L1 B4 channel."""
    filename = tm_l1_files_path / "LT04_L1TP_143021_19890818_20200916_02_T1_B4.TIF"
    create_tif_file(spectral_data, "B4", tm_l1_area, filename, tm_l1_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def tm_l1_b6_file(tm_l1_files_path, thermal_data, tm_l1_area):
    """Create the file for the Landsat TM L1 B6 channel."""
    filename = tm_l1_files_path / "LT04_L1TP_143021_19890818_20200916_02_T1_B6.TIF"
    create_tif_file(thermal_data, "B6", tm_l1_area, filename, tm_l1_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def tm_l1_sza_file(tm_l1_files_path, sza_rad_data, tm_l1_area):
    """Create the file for the Landsat TM L1 sza."""
    filename = tm_l1_files_path / "LT04_L1TP_143021_19890818_20200916_02_T1_SZA.TIF"
    create_tif_file(sza_rad_data, "sza", tm_l1_area, filename, tm_l1_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def tm_l1_mda_file(tm_l1_files_path):
    """Create the Landsat TM L1 metadata xml file."""
    filename = tm_l1_files_path / "LT04_L1TP_143021_19890818_20200916_02_T1_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "tm_l1_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def tm_l1_all_files(tm_l1_b4_file, tm_l1_b6_file, tm_l1_mda_file, tm_l1_sza_file):
    """Return all the files."""
    return tm_l1_b4_file, tm_l1_b6_file, tm_l1_mda_file, tm_l1_sza_file


@pytest.fixture(scope="module")
def tm_l2_b4_file(tm_l2_files_path, spectral_data, tm_l2_area):
    """Create the file for the Landsat TM L2 B4 channel."""
    filename = tm_l2_files_path / "LT05_L2SP_165054_20110817_20200820_02_T1_SR_B4.TIF"
    create_tif_file(spectral_data, "B4", tm_l2_area, filename, tm_l2_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def tm_l2_b6_file(tm_l2_files_path, thermal_data, tm_l2_area):
    """Create the file for the Landsat TM L2 B6 channel."""
    filename = tm_l2_files_path / "LT05_L2SP_165054_20110817_20200820_02_T1_ST_B6.TIF"
    create_tif_file(thermal_data, "B6", tm_l2_area, filename, tm_l2_date)
    return os.fspath(filename)

@pytest.fixture(scope="module")
def tm_l2_rad_file(tm_l2_files_path, sza_rad_data, tm_l2_area):
    """Create the file for the Landsat TM L2 TRAD channel."""
    filename = tm_l2_files_path / "LT05_L2SP_165054_20110817_20200820_02_T1_ST_TRAD.TIF"
    create_tif_file(sza_rad_data, "TRAD", tm_l2_area, filename, tm_l2_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def tm_l2_mda_file(tm_l2_files_path):
    """Create the Landsat TM L2 metadata xml file."""
    filename = tm_l2_files_path / "LT05_L2SP_165054_20110817_20200820_02_T1_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "tm_l2_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def tm_l2_all_files(tm_l2_b4_file, tm_l2_b6_file, tm_l2_mda_file, tm_l2_rad_file):
    """Return all the files."""
    return tm_l2_b4_file, tm_l2_b6_file, tm_l2_mda_file, tm_l2_rad_file


@pytest.fixture(scope="module")
def mss_l1_landsat1_b4_file(mss_l1_landsat1_files_path, spectral_data, mss_l1_landsat1_area):
    """Create the file for the Landsat-1 MSS L1 B4 channel."""
    filename = mss_l1_landsat1_files_path / "LM01_L1TP_032030_19720729_20200909_02_T2_B4.TIF"
    create_tif_file(spectral_data, "B4", mss_l1_landsat1_area, filename, mss_l1_landsat1_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def mss_l1_landsat1_mda_file(mss_l1_landsat1_files_path):
    """Create the Landsat-1 MSS L1 metadata xml file."""
    filename = mss_l1_landsat1_files_path / "LM01_L1TP_032030_19720729_20200909_02_T2_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "mss_l1_landsat1_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def mss_l1_landsat1_all_files(mss_l1_landsat1_b4_file, mss_l1_landsat1_mda_file):
    """Return all the files."""
    return mss_l1_landsat1_b4_file, mss_l1_landsat1_mda_file


@pytest.fixture(scope="module")
def mss_l1_landsat4_b4_file(mss_l1_landsat4_files_path, spectral_data, mss_l1_landsat4_area):
    """Create the file for the Landsat-4 MSS L1 B4 channel."""
    filename = mss_l1_landsat4_files_path / "LM04_L1TP_029030_19840415_20200903_02_T2_B4.TIF"
    create_tif_file(spectral_data, "B4", mss_l1_landsat4_area, filename, mss_l1_landsat4_date)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def mss_l1_landsat4_mda_file(mss_l1_landsat4_files_path):
    """Create the Landsat-4 MSS L1 metadata xml file."""
    filename = mss_l1_landsat4_files_path / "LM04_L1TP_029030_19840415_20200903_02_T2_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "mss_l1_landsat4_metadata.xml"), filename)
    return os.fspath(filename)


@pytest.fixture(scope="module")
def mss_l1_landsat4_all_files(mss_l1_landsat4_b4_file, mss_l1_landsat4_mda_file):
    """Return all the files."""
    return mss_l1_landsat4_b4_file, mss_l1_landsat4_mda_file


@pytest.fixture(scope="module")
def antarctic_mda_file(antarctic_files_path):
    """Create the Antarctic Landsat OLI-TIRS L1 metadata xml file."""
    filename = antarctic_files_path / "LC08_L1GT_132122_20220301_20220308_02_T2_MTL.xml"
    shutil.copyfile(os.path.join(ETC_DIR, "antarctic_metadata.xml"), filename)
    return os.fspath(filename)


def get_filename_info(date, level_correction, spacecraft, data_type):
    """Set up a filename info dict."""
    return dict(observation_date=date,
                platform_type="L",
                process_level_correction=level_correction,
                spacecraft_id=spacecraft,
                data_type=data_type,
                collection_id="02")


class TestLandsat:
    """Test Landsat image readers."""

    ftype_info = {"file_type": "granule_B4"}

    @pytest.mark.parametrize(
        ("reader", "spectral_name", "thermal_name", "all_files", "area"),
        [
            pytest.param(
                "oli_tirs_l1_tif", "B4", "B11", lf("oli_tirs_l1_all_files"), lf("oli_tirs_l1_area"), id="oli_tirs_l1",
            ),
            pytest.param(
                "oli_tirs_l2_tif", "B4", "B10", lf("oli_tirs_l2_all_files"), lf("oli_tirs_l2_area"), id="oli_tirs_l2",
            ),
            pytest.param("etm_l1_tif", "B4", "B6_VCID_1", lf("etm_l1_all_files"), lf("etm_l1_area"), id="etm_l1"),
            pytest.param("etm_l2_tif", "B4", "B6", lf("etm_l2_all_files"), lf("etm_l2_area"), id="etm_l2"),
            pytest.param("tm_l1_tif", "B4", "B6", lf("tm_l1_all_files"), lf("tm_l1_area"), id="tm_l1"),
            pytest.param("tm_l2_tif", "B4", "B6", lf("tm_l2_all_files"), lf("tm_l2_area"), id="tm_l2"),
            pytest.param(
                "mss_l1_tif", "B4", None, lf("mss_l1_landsat1_all_files"), lf("mss_l1_landsat1_area"),
                id="mss_l1_landsat1",
            ),
            pytest.param(
                "mss_l1_tif", "B4", None, lf("mss_l1_landsat4_all_files"), lf("mss_l1_landsat4_area"),
                id="mss_l1_landsat4",
            ),
        ],
    )
    @pytest.mark.parametrize("remote", [True, False])
    def test_basicload(self, reader, area, all_files, spectral_name, thermal_name, remote):
        """Test loading a Landsat Scene."""
        if remote:
            all_files = convert_to_fsfile(all_files)

        scn = Scene(reader=reader, filenames=all_files)
        if thermal_name is not None:
            scn.load([spectral_name, thermal_name])
        else:
            scn.load([spectral_name])

        self._check_basicload_thermal(scn, spectral_name, thermal_name, area, reader)
        self._check_basicload_mss_l1_tif(reader, scn)


    @staticmethod
    def _check_basicload_thermal(scn, spectral_name, thermal_name, area, reader):
        # Check dataset is loaded correctly
        assert scn[spectral_name].shape == (100, 100)
        assert scn[spectral_name].attrs["area"] == area
        assert scn[spectral_name].attrs["saturated"]
        if thermal_name is not None:
            assert scn[thermal_name].shape == (100, 100)
            assert scn[thermal_name].attrs["area"] == area
            if reader in ["oli_tirs_l1_tif", "oli_tirs_l2_tif", "etm_l2_tif"]:
                # OLI TIRS and ETM+ L2 do not have saturation flag on thermal band
                with pytest.raises(KeyError, match="saturated"):
                    assert not scn[thermal_name].attrs["saturated"]
            else:
                assert not scn[thermal_name].attrs["saturated"]


    @staticmethod
    def _check_basicload_mss_l1_tif(reader, scn):
        # Check if MSS sets wavelength correctly
        if reader == "mss_l1_tif":
            if scn["B4"].attrs["platform_name"] == "Landsat-4":
                assert scn["B4"].attrs["_satpy_id"]["wavelength"].min == 0.8
                assert scn["B4"].attrs["_satpy_id"]["wavelength"].central == 0.95
                assert scn["B4"].attrs["_satpy_id"]["wavelength"].max == 1.1
            elif scn["B4"].attrs["platform_name"] == "Landsat-1":
                assert scn["B4"].attrs["_satpy_id"]["wavelength"].min == 0.5
                assert scn["B4"].attrs["_satpy_id"]["wavelength"].central == 0.55
                assert scn["B4"].attrs["_satpy_id"]["wavelength"].max == 0.6


    @pytest.mark.parametrize(
        ("reader", "date_time", "spectral_file", "spectral_name", "mda_file"),
        [
            pytest.param(
                "oli_tirs_l1_tif", oli_tirs_l1_datetime, lf("oli_tirs_l1_b4_file"), "B4", lf("oli_tirs_l1_mda_file"),
                id="oli_tirs_l1",
            ),
            pytest.param(
                "oli_tirs_l2_tif", oli_tirs_l2_datetime, lf("oli_tirs_l2_b4_file"), "B4", lf("oli_tirs_l2_mda_file"),
                id="oli_tirs_l2",
            ),
            pytest.param("etm_l1_tif", etm_l1_datetime, lf("etm_l1_b4_file"), "B4", lf("etm_l1_mda_file"), id="etm_l1"),
            pytest.param("etm_l2_tif", etm_l2_datetime, lf("etm_l2_b4_file"), "B4", lf("etm_l2_mda_file"), id="etm_l2"),
            pytest.param("tm_l1_tif", tm_l1_datetime, lf("tm_l1_b4_file"), "B4", lf("tm_l1_mda_file"), id="tm_l1"),
            pytest.param("tm_l2_tif", tm_l2_datetime, lf("tm_l2_b4_file"), "B4", lf("tm_l2_mda_file"), id="tm_l2"),
            pytest.param(
                "mss_l1_tif", mss_l1_landsat1_datetime, lf("mss_l1_landsat1_b4_file"), "B4",
                lf("mss_l1_landsat1_mda_file"), id="mss_l1_landsat1",
            ),
            pytest.param(
                "mss_l1_tif", mss_l1_landsat4_datetime, lf("mss_l1_landsat4_b4_file"), "B4",
                lf("mss_l1_landsat4_mda_file"), id="mss_l1_landsat4",
            ),
        ],
    )
    def test_ch_startend(self, reader, spectral_file, mda_file, spectral_name, date_time):
        """Test correct retrieval of start/end times."""
        scn = Scene(reader=reader, filenames=[spectral_file, mda_file])
        bnds = scn.available_dataset_names()
        assert bnds == [spectral_name]

        scn.load(["B4"])
        assert scn.start_time == date_time
        assert scn.end_time == date_time

    @pytest.mark.parametrize(
        (
            "spectral_file", "mda_file", "CH_reader_class", "MD_reader_class",
            "filename_info",
        ),
        [
            pytest.param(
                lf("oli_tirs_l1_b4_file"), lf("oli_tirs_l1_mda_file"), OLITIRSCHReader, LandsatL1MDReader,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"), id="oli_tirs_l1",
            ),
            pytest.param(
                lf("oli_tirs_l2_b4_file"), lf("oli_tirs_l2_mda_file"), OLITIRSL2CHReader, LandsatL2MDReader,
                get_filename_info(oli_tirs_l2_date, "L2SP", "09", "C"), id="oli_tirs_l2",
            ),
            pytest.param(
               lf( "etm_l1_b4_file"), lf("etm_l1_mda_file"), ETMCHReader, LandsatL1MDReader,
                get_filename_info(etm_l1_date, "L1TP", "07", "E"), id="etm_l1",
            ),
            pytest.param(
                lf("etm_l2_b4_file"), lf("etm_l2_mda_file"), ETML2CHReader, LandsatL2MDReader,
                get_filename_info(etm_l2_date, "L2SP", "07", "E"), id="etm_l2",
            ),
            pytest.param(
                lf("tm_l1_b4_file"), lf("tm_l1_mda_file"), TMCHReader, LandsatL1MDReader,
                get_filename_info(tm_l1_date, "L1TP", "04", "T"), id="tm_l1",
            ),
            pytest.param(
                lf("tm_l2_b4_file"), lf("tm_l2_mda_file"), TML2CHReader, LandsatL2MDReader,
                get_filename_info(tm_l2_date, "L2SP", "05", "T"), id="tm_l2",
            ),
            pytest.param(
                lf("mss_l1_landsat1_b4_file"), lf("mss_l1_landsat1_mda_file"), MSSCHReader, LandsatL1MDReader,
                get_filename_info(mss_l1_landsat1_date, "L1TP", "01", "M"), id="mss_l1_landsat1",
            ),
            pytest.param(
                lf("mss_l1_landsat4_b4_file"), lf("mss_l1_landsat4_mda_file"), MSSCHReader, LandsatL1MDReader,
                get_filename_info(mss_l1_landsat4_date, "L1TP", "04", "M"), id="mss_l1_landsat4",
            ),
        ],
    )
    def test_loading_gd(
        self,
        CH_reader_class,
        MD_reader_class,
        mda_file,
        spectral_file,
        filename_info,
    ):
        """Test loading a Landsat Scene with good channel requests."""
        good_mda = MD_reader_class(mda_file, filename_info, {})
        rdr = CH_reader_class(spectral_file, filename_info, self.ftype_info, good_mda)

        # Check case with good file data and load request
        rdr.get_dataset({"name": "B4", "calibration": "counts"}, {"standard_name": "test_data", "units": "test_units"})

    @pytest.mark.parametrize(
        (
            "spectral_file", "mda_file", "CH_reader_class", "MD_reader_class",
            "filename_info",
        ),
        [
            pytest.param(
                lf("oli_tirs_l1_b4_file"), lf("oli_tirs_l1_mda_file"), OLITIRSCHReader, LandsatL1MDReader,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"), id="oli_tirs_l1",
            ),
            pytest.param(
                lf("oli_tirs_l2_b4_file"), lf("oli_tirs_l2_mda_file"), OLITIRSL2CHReader, LandsatL2MDReader,
                get_filename_info(oli_tirs_l2_date, "L2SP", "09", "C"), id="oli_tirs_l2",
            ),
            pytest.param(
                lf("etm_l1_b4_file"), lf("etm_l1_mda_file"), ETMCHReader, LandsatL1MDReader,
                get_filename_info(etm_l1_date, "L1TP", "07", "E"), id="etm_l1",
            ),
            pytest.param(
                lf("etm_l2_b4_file"), lf("etm_l2_mda_file"), ETML2CHReader, LandsatL2MDReader,
                get_filename_info(etm_l2_date, "L2SP", "07", "E"), id="etm_l2",
            ),
            pytest.param(
                lf("tm_l1_b4_file"), lf("tm_l1_mda_file"), TMCHReader, LandsatL1MDReader,
                get_filename_info(tm_l1_date, "L1TP", "04", "T"), id="tm_l1",
            ),
            pytest.param(
                lf("tm_l2_b4_file"), lf("tm_l2_mda_file"), TML2CHReader, LandsatL2MDReader,
                get_filename_info(tm_l2_date, "L2SP", "05", "T"), id="tm_l2",
            ),
            pytest.param(
                lf("mss_l1_landsat1_b4_file"), lf("mss_l1_landsat1_mda_file"), MSSCHReader, LandsatL1MDReader,
                get_filename_info(mss_l1_landsat1_date, "L1TP", "01", "M"), id="mss_l1_landsat1",
            ),
            pytest.param(
                lf("mss_l1_landsat4_b4_file"), lf("mss_l1_landsat4_mda_file"), MSSCHReader, LandsatL1MDReader,
                get_filename_info(mss_l1_landsat4_date, "L1TP", "04", "M"), id="mss_l1_landsat4",
            ),
        ],
    )
    def test_loading_badfil(
        self,
        CH_reader_class,
        MD_reader_class,
        mda_file,
        spectral_file,
        filename_info,
    ):
        """Test loading a Landsat Scene with bad channel requests."""
        good_mda = MD_reader_class(mda_file, filename_info, {})
        rdr = CH_reader_class(spectral_file, filename_info, self.ftype_info, good_mda)

        ftype = {"standard_name": "test_data", "units": "test_units"}
        # Check case with request to load channel not matching filename
        with pytest.raises(ValueError, match="Requested channel B5 does not match the reader channel B4"):
            rdr.get_dataset({"name": "B5", "calibration": "counts"}, ftype)

    @pytest.mark.parametrize(
        (
            "thermal_file", "mda_file", "CH_reader_class", "MD_reader_class",
            "filename_info",
        ),
        [
            pytest.param(
                lf("oli_tirs_l1_b11_file"), lf("oli_tirs_l1_mda_file"), OLITIRSCHReader, LandsatL1MDReader,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"), id="oli_tirs_l1",
            ),
            pytest.param(
                lf("oli_tirs_l2_b10_file"), lf("oli_tirs_l2_mda_file"), OLITIRSL2CHReader, LandsatL2MDReader,
                get_filename_info(oli_tirs_l2_date, "L2SP", "09", "C"), id="oli_tirs_l2",
            ),
        ],
    )
    def test_loading_badchan(
        self,
        CH_reader_class,
        MD_reader_class,
        mda_file,
        thermal_file,
        filename_info,
    ):
        """Test loading a Landsat Scene with bad channel requests."""
        good_mda = MD_reader_class(mda_file, filename_info, {})
        ftype = {"standard_name": "test_data", "units": "test_units"}
        bad_finfo = filename_info.copy()
        bad_finfo["data_type"] = "T"

        # Check loading invalid channel for data type
        rdr = CH_reader_class(thermal_file, bad_finfo, self.ftype_info, good_mda)
        with pytest.raises(ValueError, match="Requested channel B4 is not available in this granule"):
            rdr.get_dataset({"name": "B4", "calibration": "counts"}, ftype)

        if filename_info["process_level_correction"] == "L1TP":
            # L1 test
            bad_finfo["data_type"] = "O"
            ftype_b11 = self.ftype_info.copy()
            ftype_b11["file_type"] = "granule_B11"
            rdr = CH_reader_class(thermal_file, bad_finfo, ftype_b11, good_mda)
            with pytest.raises(ValueError, match="Requested channel B11 is not available in this granule"):
                rdr.get_dataset({"name": "B11", "calibration": "counts"}, ftype)
        else:
            # L2 test
            bad_finfo["data_type"] = "O"
            ftype_b10 = self.ftype_info.copy()
            ftype_b10["file_type"] = "granule_B10"
            rdr = OLITIRSL2CHReader(thermal_file, bad_finfo, ftype_b10, good_mda)
            with pytest.raises(ValueError, match="Requested channel B10 is not available in this granule"):
                rdr.get_dataset({"name": "B10", "calibration": "counts"}, ftype)

    @pytest.mark.parametrize(
        (
            "spectral_file", "mda_file", "CH_reader_class", "MD_reader_class",
            "filename_info",
        ),
        [
            pytest.param(
                lf("oli_tirs_l1_b4_file"), lf("oli_tirs_l1_mda_file"), OLITIRSCHReader, LandsatL1MDReader,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"), id="oli_tirs_l1",
            ),
            pytest.param(
                lf("oli_tirs_l2_b4_file"), lf("oli_tirs_l2_mda_file"), OLITIRSL2CHReader, LandsatL2MDReader,
                get_filename_info(oli_tirs_l2_date, "L2SP", "09", "C"), id="oli_tirs_l2",
            ),
            pytest.param(
                lf("etm_l1_b4_file"), lf("etm_l1_mda_file"), ETMCHReader, LandsatL1MDReader,
                get_filename_info(etm_l1_date, "L1TP", "07", "E"), id="etm_l1",
            ),
            pytest.param(
                lf("etm_l2_b4_file"), lf("etm_l2_mda_file"), ETML2CHReader, LandsatL2MDReader,
                get_filename_info(etm_l2_date, "L2SP", "07", "E"), id="etm_l2",
            ),
            pytest.param(
                lf("tm_l1_b4_file"), lf("tm_l1_mda_file"), TMCHReader, LandsatL1MDReader,
                get_filename_info(tm_l1_date, "L1TP", "04", "T"), id="tm_l1",
            ),
            pytest.param(
                lf("tm_l2_b4_file"), lf("tm_l2_mda_file"), TML2CHReader, LandsatL2MDReader,
                get_filename_info(tm_l2_date, "L2SP", "05", "T"), id="tm_l2",
            ),
            pytest.param(
                lf("mss_l1_landsat1_b4_file"), lf("mss_l1_landsat1_mda_file"), MSSCHReader, LandsatL1MDReader,
                get_filename_info(mss_l1_landsat1_date, "L1TP", "01", "M"), id="mss_l1_landsat1",
            ),
            pytest.param(
                lf("mss_l1_landsat4_b4_file"), lf("mss_l1_landsat4_mda_file"), MSSCHReader, LandsatL1MDReader,
                get_filename_info(mss_l1_landsat4_date, "L1TP", "04", "M"), id="mss_l1_landsat4",
            ),
        ],
    )
    def test_badfiles(
        self,
        CH_reader_class,
        MD_reader_class,
        mda_file,
        spectral_file,
        filename_info,
    ):
        """Test loading a Landsat Scene with bad data."""
        bad_fname_info = filename_info.copy()
        bad_fname_info["platform_type"] = "B"

        bad_ftype_info = self.ftype_info.copy()
        bad_ftype_info["file_type"] = "granule-b05"

        ftype = {"standard_name": "test_data", "units": "test_units"}

        # Test that metadata reader initialises with correct filename
        good_mda = MD_reader_class(mda_file, filename_info, ftype)

        # Check metadata reader fails if platform type is wrong
        with pytest.raises(ValueError, match="This reader only supports Landsat data"):
            MD_reader_class(mda_file, bad_fname_info, ftype)

        # Test that metadata reader initialises with correct filename
        CH_reader_class(spectral_file, filename_info, self.ftype_info, good_mda)

        # Check metadata reader fails if platform type is wrong
        with pytest.raises(ValueError, match="This reader only supports Landsat data"):
            CH_reader_class(spectral_file, bad_fname_info, self.ftype_info, good_mda)

        with pytest.raises(ValueError, match="Invalid file type: granule-b05"):
            CH_reader_class(spectral_file, filename_info, bad_ftype_info, good_mda)

    @pytest.mark.parametrize(
        ("reader", "spectral_name", "thermal_name", "all_files"),
        [
            pytest.param("oli_tirs_l1_tif", "B4", "B11", lf("oli_tirs_l1_all_files"), id="oli_tirs_l1"),
            pytest.param("oli_tirs_l2_tif", "B4", "B10", lf("oli_tirs_l2_all_files"), id="oli_tirs_l2"),
            pytest.param("etm_l1_tif", "B4", "B6_VCID_1", lf("etm_l1_all_files"), id="etm_l1"),
            pytest.param("etm_l2_tif", "B4", "B6", lf("etm_l2_all_files"), id="etm_l2"),
            pytest.param("tm_l1_tif", "B4", "B6", lf("tm_l1_all_files"), id="tm_l1"),
            pytest.param("tm_l2_tif", "B4", "B6", lf("tm_l2_all_files"), id="tm_l2"),
            pytest.param("mss_l1_tif", "B4", None, lf("mss_l1_landsat1_all_files"), id="mss_l1_landsat1"),
            pytest.param("mss_l1_tif", "B4", None, lf("mss_l1_landsat4_all_files"), id="mss_l1_landsat4"),
        ],
    )
    def test_calibration_counts(
        self,
        reader,
        all_files,
        spectral_name,
        thermal_name,
        spectral_data,
        thermal_data,
    ):
        """Test counts calibration mode for the reader."""
        scn = Scene(reader=reader, filenames=all_files)
        if thermal_name is not None:
            scn.load([spectral_name, thermal_name], calibration="counts")
        else:
            scn.load([spectral_name], calibration="counts")

        np.testing.assert_allclose(scn[spectral_name].values, spectral_data)
        assert scn[spectral_name].attrs["units"] == "1"
        assert scn[spectral_name].attrs["standard_name"] == "counts"

        if thermal_name is not None:
            np.testing.assert_allclose(scn[thermal_name].values, thermal_data)
            assert scn[thermal_name].attrs["units"] == "1"
            assert scn[thermal_name].attrs["standard_name"] == "counts"

    @pytest.mark.parametrize(
        ("reader", "spectral_name", "thermal_name", "all_files", "cal_spectral_params", "cal_thermal_params"),
        [
            pytest.param(
                "oli_tirs_l1_tif", "B4", "B11", lf("oli_tirs_l1_all_files"),
                oli_tirs_l1_sp_params, oli_tirs_l1_th_params, id="oli_tirs_l1",
            ),
            pytest.param(
                "etm_l1_tif", "B4", "B6_VCID_1", lf("etm_l1_all_files"), etm_l1_sp_params, etm_l1_th_params,
                id="etm_l1",
            ),
            pytest.param("tm_l1_tif", "B4", "B6", lf("tm_l1_all_files"), tm_l1_sp_params, tm_l1_th_params,
                         id="tm_l1",
                         ),
            pytest.param(
                "mss_l1_tif", "B4", None, lf("mss_l1_landsat1_all_files"), mss_l1_landsat1_sp_params, None,
                id="mss_l1_landsat1",
            ),
            pytest.param(
                "mss_l1_tif", "B4", None, lf("mss_l1_landsat4_all_files"), mss_l1_landsat4_sp_params, None,
                id="mss_l1_landsat4",
            ),
        ],
    )
    def test_calibration_radiance(
        self,
        reader,
        all_files,
        spectral_name,
        thermal_name,
        cal_spectral_params,
        cal_thermal_params,
        spectral_data,
        thermal_data,
    ):
        """Test radiance calibration mode for the reader."""
        exp_spectral = (spectral_data * cal_spectral_params[0] + cal_spectral_params[1]).astype(np.float32)
        if thermal_name is not None:
            exp_thermal = (thermal_data * cal_thermal_params[0] + cal_thermal_params[1]).astype(np.float32)

        scn = Scene(reader=reader, filenames=all_files)
        if thermal_name is not None:
            scn.load([spectral_name, thermal_name], calibration="radiance")
        else:
            scn.load([spectral_name], calibration="radiance")

        assert scn[spectral_name].attrs["units"] == "W m-2 um-1 sr-1"
        assert scn[spectral_name].attrs["standard_name"] == "toa_outgoing_radiance_per_unit_wavelength"
        np.testing.assert_allclose(scn[spectral_name].values, exp_spectral, rtol=1e-4)

        if thermal_name is not None:
            assert scn[thermal_name].attrs["units"] == "W m-2 um-1 sr-1"
            assert scn[thermal_name].attrs["standard_name"] == "toa_outgoing_radiance_per_unit_wavelength"
            np.testing.assert_allclose(scn[thermal_name].values, exp_thermal, rtol=1e-4)

    @pytest.mark.parametrize(
        (
            "reader", "spectral_name", "thermal_name", "sza_rad_name", "all_files",
            "cal_spectral_params", "cal_thermal_params",
        ),
        [
            pytest.param(
                "oli_tirs_l1_tif", "B4", "B11", None, lf("oli_tirs_l1_all_files"),
                oli_tirs_l1_sp_params, oli_tirs_l1_th_params, id="oli_tirs_l1",
            ),
            pytest.param(
                "oli_tirs_l2_tif", "B4", "B10", "TRAD", lf("oli_tirs_l2_all_files"),
                oli_tirs_l2_sp_params, oli_tirs_l2_th_params, id="oli_tirs_l2",
            ),
            pytest.param(
                "etm_l1_tif", "B4", "B6_VCID_1", None, lf("etm_l1_all_files"),
                etm_l1_sp_params, etm_l1_th_params, id="etm_l1",
            ),
            pytest.param(
                "etm_l2_tif", "B4", "B6", "TRAD", lf("etm_l2_all_files"), etm_l2_sp_params, etm_l2_th_params,
                id="etm_l2",
            ),
            pytest.param(
                "tm_l1_tif", "B4", "B6", None, lf("tm_l1_all_files"), tm_l1_sp_params, tm_l1_th_params,
                id="tm_l1",
            ),
            pytest.param(
                "tm_l2_tif", "B4", "B6", "TRAD", lf("tm_l2_all_files"), tm_l2_sp_params, tm_l2_th_params,
                id="tm_l2",
            ),
            pytest.param(
                "mss_l1_tif", "B4", None, None, lf("mss_l1_landsat1_all_files"),
                mss_l1_landsat1_sp_params, None, id="mss_l1_landsat1",
            ),
            pytest.param(
                "mss_l1_tif", "B4", None, None, lf("mss_l1_landsat4_all_files"),
                mss_l1_landsat4_sp_params, None, id="mss_l1_landsat4",
            ),
        ],
    )
    def test_calibration_highlevel(
        self,
        reader,
        all_files,
        spectral_name,
        thermal_name,
        sza_rad_name,
        cal_spectral_params,
        cal_thermal_params,
        spectral_data,
        thermal_data,
        sza_rad_data,
    ):
        """Test high level calibration modes for the reader."""
        exp_spectral = self._get_expected_highlevel_spectral(reader, spectral_data, cal_spectral_params)
        exp_thermal, exp_rad = self._get_expected_highlevel_thermal(
            reader, thermal_data, cal_thermal_params, sza_rad_data, thermal_name)

        scn = self._get_scn_highlevel(reader, all_files, spectral_name, thermal_name, sza_rad_name)

        assert scn[spectral_name].attrs["units"] == "%"
        if "_l2_" in reader:
            assert scn[spectral_name].attrs["standard_name"] == "surface_bidirectional_reflectance"
        else:
            assert scn[spectral_name].attrs["standard_name"] == "toa_bidirectional_reflectance"
        np.testing.assert_allclose(np.array(scn[spectral_name].values), np.array(exp_spectral), rtol=1e-4)

        if thermal_name is not None:
            self._check_thermal_highlevel(scn, thermal_name, exp_thermal)
            self._check_thermal_l2_highlevel(reader, scn, sza_rad_name, exp_rad)

    @staticmethod
    def _get_expected_highlevel_spectral(reader, spectral_data, cal_spectral_params):
        if "_l2_" in reader:
            return (spectral_data * cal_spectral_params[0] + cal_spectral_params[1]).astype(np.float32) * 100
        return (spectral_data * cal_spectral_params[2] + cal_spectral_params[3]).astype(np.float32) * 100

    @staticmethod
    def _get_expected_highlevel_thermal(reader, thermal_data, cal_thermal_params, sza_rad_data, thermal_name):
        exp_thermal, exp_rad = None, None
        if "_l2_" in reader and thermal_name is not None:
            exp_thermal = (thermal_data * cal_thermal_params[0] + cal_thermal_params[1]).astype(np.float32)
            exp_rad = (sza_rad_data * 0.001).astype(np.float32)
        elif thermal_name is not None:
            exp_thermal = (thermal_data * cal_thermal_params[0] + cal_thermal_params[1])
            exp_thermal = (cal_thermal_params[3] / np.log((cal_thermal_params[2] / exp_thermal) + 1))
            exp_thermal = exp_thermal.astype(np.float32)
        return exp_thermal, exp_rad

    @staticmethod
    def _get_scn_highlevel(reader, all_files, spectral_name, thermal_name, sza_rad_name):
        scn = Scene(reader=reader, filenames=all_files)
        if thermal_name is not None:
            if "_l2_" in reader:
                scn.load([spectral_name, thermal_name, sza_rad_name])
            else:
                scn.load([spectral_name, thermal_name])
        else:
            scn.load([spectral_name])
        return scn

    @staticmethod
    def _check_thermal_highlevel(scn, thermal_name, exp_thermal):
        assert scn[thermal_name].attrs["units"] == "K"
        assert scn[thermal_name].attrs["standard_name"] == "brightness_temperature"
        np.testing.assert_allclose(scn[thermal_name].values, exp_thermal, rtol=1e-6)

    @staticmethod
    def _check_thermal_l2_highlevel(reader, scn, sza_rad_name, exp_rad):
        if "_l2_" in reader:
            assert scn[sza_rad_name].attrs["units"] == "W m-2 um-1 sr-1"
            assert scn[sza_rad_name].attrs["standard_name"] == "toa_outgoing_radiance_per_unit_wavelength"
            np.testing.assert_allclose(scn[sza_rad_name].values, exp_rad, rtol=1e-6)

    @pytest.mark.parametrize(
        ("reader", "all_files"),
        [
            pytest.param("oli_tirs_l1_tif", lf("oli_tirs_l1_all_files"), id="oli_tirs_l1"),
            pytest.param("etm_l1_tif", lf("etm_l1_all_files"), id="etm_l1",),
            pytest.param("tm_l1_tif", lf("tm_l1_all_files"), id="tm_l1"),
        ],
    )
    def test_angles(self, reader, all_files, sza_rad_data):
        """Test calibration modes for the reader."""
        # Check angles are calculated correctly
        scn = Scene(reader=reader, filenames=all_files)
        scn.load(["solar_zenith_angle"])
        assert scn["solar_zenith_angle"].attrs["units"] == "degrees"
        assert scn["solar_zenith_angle"].attrs["standard_name"] == "solar_zenith_angle"
        np.testing.assert_allclose(scn["solar_zenith_angle"].values * 100,
                                   np.array(sza_rad_data),
                                   atol=0.01,
                                   rtol=1e-3)

    @pytest.mark.parametrize(
        (
            "reader", "MD_reader_class", "mda_file", "cal_dict",
            "filename_info", "platform_name", "earth_sun_distance",
        ),
        [
            pytest.param(
                "oli_tirs_l1_tif", LandsatL1MDReader, lf("oli_tirs_l1_mda_file"), oli_tirs_l1_cal_dict,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"),
                "Landsat-8", 1.0079981, id="oli_tirs_l1",
            ),
            pytest.param(
                "oli_tirs_l2_tif", LandsatL2MDReader, lf("oli_tirs_l2_mda_file"), oli_tirs_l2_cal_dict,
                get_filename_info(oli_tirs_l2_date, "L2SP", "09", "C"),
                "Landsat-9", 1.0158933, id="oli_tirs_l2",
            ),
            pytest.param(
                "etm_l1_tif", LandsatL1MDReader, lf("etm_l1_mda_file"), etm_l1_cal_dict,
                get_filename_info(etm_l1_date, "L1TP", "07", "E"),
                "Landsat-7", 0.9850987, id="etm_l1",
            ),
            pytest.param(
                "etm_l2_tif", LandsatL2MDReader, lf("etm_l2_mda_file"), etm_l2_cal_dict,
                get_filename_info(etm_l2_date, "L2SP", "07", "E"),
                "Landsat-7", 1.0124651, id="etm_l2",
            ),
            pytest.param(
                "tm_l1_tif", LandsatL1MDReader, lf("tm_l1_mda_file"), tm_l1_cal_dict,
                get_filename_info(tm_l1_date, "L1TP", "04", "T"),
                "Landsat-4", 1.0122057, id="tm_l1",
            ),
            pytest.param(
                "tm_l2_tif", LandsatL2MDReader, lf("tm_l2_mda_file"), tm_l2_cal_dict,
                get_filename_info(tm_l2_date, "L2SP", "05", "T"),
                "Landsat-5", 1.0125021, id="tm_l2",
            ),
            pytest.param(
                "mss_l1_tif", LandsatL1MDReader, lf("mss_l1_landsat1_mda_file"), mss_l1_landsat1_cal_dict,
                get_filename_info(mss_l1_landsat1_date, "L1TP", "01", "M"),
                "Landsat-1", 1.0152109, id="mss_l1_landsat1",
            ),
            pytest.param(
                "mss_l1_tif", LandsatL1MDReader, lf("mss_l1_landsat4_mda_file"), mss_l1_landsat4_cal_dict,
                get_filename_info(mss_l1_landsat4_date, "L1TP", "04", "M"),
                "Landsat-4", 1.0035512, id="mss_l1_landsat4",
            ),
        ],
    )
    def test_metadata(
        self,
        reader,
        MD_reader_class,
        mda_file,
        cal_dict,
        platform_name,
        earth_sun_distance,
        filename_info,
    ):
        """Check that metadata values loaded correctly."""
        mda = MD_reader_class(mda_file, filename_info, {})

        assert mda.platform_name == platform_name
        assert mda.earth_sun_distance() == earth_sun_distance

        cal_bands = list(cal_dict.keys())
        assert mda.band_calibration[cal_bands[0]] == cal_dict[cal_bands[0]]
        assert mda.band_calibration[cal_bands[1]] == cal_dict[cal_bands[1]]
        assert mda.band_calibration[cal_bands[2]] == cal_dict[cal_bands[2]]
        assert not mda.band_saturation[cal_bands[0]]
        assert mda.band_saturation["B4"]
        assert not mda.band_saturation[cal_bands[1]]
        if reader in ["oli_tirs_l1_tif", "oli_tirs_l2_tif", "etm_l2_tif"]:
            with pytest.raises(KeyError):
                mda.band_saturation[cal_bands[2]]
        else:
            assert not mda.band_saturation[cal_bands[2]]

    @pytest.mark.parametrize(
        (
            "MD_reader_class", "mda_file", "extent", "pan_extent",
            "filename_info",
        ),
        [
            pytest.param(
                LandsatL1MDReader, lf("oli_tirs_l1_mda_file"), oli_tirs_l1_extent, oli_tirs_l1_pan_extent,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"), id="oli_tirs_l1",
            ),
            pytest.param(
                LandsatL2MDReader, lf("oli_tirs_l2_mda_file"), oli_tirs_l2_extent, None,
                get_filename_info(oli_tirs_l2_date, "L2SP", "09", "C"), id="oli_tirs_l2",
            ),
            pytest.param(
                LandsatL1MDReader, lf("etm_l1_mda_file"), etm_l1_extent, etm_l1_pan_extent,
                get_filename_info(etm_l1_date, "L1TP", "07", "E"), id="etm_l1",
            ),
            pytest.param(
                LandsatL2MDReader, lf("etm_l2_mda_file"), etm_l2_extent, None,
                get_filename_info(etm_l2_date, "L2SP", "07", "E"), id="etm_l2",
            ),
            pytest.param(
                LandsatL1MDReader, lf("tm_l1_mda_file"), tm_l1_extent, None,
                get_filename_info(tm_l1_date, "L1TP", "04", "T"), id="tm_l1",
            ),
            pytest.param(
                LandsatL2MDReader, lf("tm_l2_mda_file"), tm_l2_extent, None,
                get_filename_info(tm_l2_date, "L2SP", "05", "T"), id="tm_l2",
            ),
            pytest.param(
                LandsatL1MDReader, lf("mss_l1_landsat1_mda_file"), mss_l1_landsat1_extent, None,
                get_filename_info(mss_l1_landsat1_date, "L1TP", "01", "M"), id="mss_l1_landsat1",
            ),
            pytest.param(
                LandsatL1MDReader, lf("mss_l1_landsat4_mda_file"), mss_l1_landsat4_extent, None,
                get_filename_info(mss_l1_landsat4_date, "L1TP", "04", "M"), id="mss_l1_landsat4",
            ),
            pytest.param(
                LandsatL1MDReader, lf("antarctic_mda_file"), antarctic_extent, antarctic_pan_extent,
                get_filename_info(oli_tirs_l1_date, "L1TP", "08", "C"), id="antarctic",
            ),
        ],
    )
    def test_area_def(
        self,
        MD_reader_class,
        mda_file,
        extent,
        pan_extent,
        filename_info,
    ):
        """Check we can get the area defs properly."""
        mda = MD_reader_class(mda_file, filename_info, {})
        standard_area = mda.build_area_def("B4")
        assert standard_area.area_extent == extent

        if pan_extent is not None:
            pan_area = mda.build_area_def("B8")
            assert pan_area.area_extent == pan_extent



"""EWA resamplers."""

from pyresample.ewa import DaskEWAResampler, LegacyDaskEWAResampler


def get_resampler_classes():
    """Get bucket resampler classes."""
    return {
        "ewa": DaskEWAResampler,
        "ewa_legacy": LegacyDaskEWAResampler,
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Demo data download helper functions for ABI L1b data."""
import os

from satpy import config


def get_us_midlatitude_cyclone_abi(base_dir=None, method=None, force=False):
    """Get GOES-16 ABI (CONUS sector) data from 2019-03-14 00:00Z.

    Args:
        base_dir (str): Base directory for downloaded files.
        method (str): Force download method for the data if not already cached.
            Allowed options are: 'gcsfs'. Default of ``None`` will
            choose the best method based on environment settings.
        force (bool): Force re-download of data regardless of its existence on
            the local system. Warning: May delete non-demo files stored in
            download directory.

    Total size: ~110MB

    """
    base_dir = base_dir or config.get("demo_data_dir", ".")
    if method is None:
        method = "gcsfs"
    if method not in ["gcsfs"]:
        raise NotImplementedError("Demo data download method '{}' not "
                                  "implemented yet.".format(method))

    from ._google_cloud_platform import get_bucket_files
    patterns = ["gs://gcp-public-data-goes-16/ABI-L1b-RadC/2019/073/00/*s20190730002*.nc"]
    subdir = os.path.join(base_dir, "abi_l1b", "20190314_us_midlatitude_cyclone")
    os.makedirs(subdir, exist_ok=True)
    filenames = get_bucket_files(patterns, subdir, force=force)
    if len(filenames) != 16:
        raise RuntimeError("Not all files could be downloaded")
    return filenames


def get_hurricane_florence_abi(base_dir=None, method=None, force=False,
                               channels=None, num_frames=10):
    """Get GOES-16 ABI (Meso sector) data from 2018-09-11 13:00Z to 17:00Z.

    Args:
        base_dir (str): Base directory for downloaded files.
        method (str): Force download method for the data if not already cached.
            Allowed options are: 'gcsfs'. Default of ``None`` will
            choose the best method based on environment settings.
        force (bool): Force re-download of data regardless of its existence on
            the local system. Warning: May delete non-demo files stored in
            download directory.
        channels (list): Channels to include in download. Defaults to all
            16 channels.
        num_frames (int or slice): Number of frames to download. Maximum
            240 frames. Default 10 frames.

    Size per frame (all channels): ~15MB

    Total size (default 10 frames, all channels): ~124MB

    Total size (240 frames, all channels): ~3.5GB

    """
    base_dir = base_dir or config.get("demo_data_dir", ".")
    if channels is None:
        channels = range(1, 17)
    if method is None:
        method = "gcsfs"
    if method not in ["gcsfs"]:
        raise NotImplementedError("Demo data download method '{}' not "
                                  "implemented yet.".format(method))
    if isinstance(num_frames, (int, float)):
        frame_slice = slice(0, num_frames)
    else:
        frame_slice = num_frames

    from ._google_cloud_platform import get_bucket_files

    patterns = []
    for channel in channels:
        # patterns += ['gs://gcp-public-data-goes-16/ABI-L1b-RadM/2018/254/1[3456]/'
        #              '*C{:02d}*s20182541[3456]*.nc'.format(channel)]
        patterns += [(
            "gs://gcp-public-data-goes-16/ABI-L1b-RadM/2018/254/13/*RadM1*C{:02d}*s201825413*.nc".format(channel),
            "gs://gcp-public-data-goes-16/ABI-L1b-RadM/2018/254/14/*RadM1*C{:02d}*s201825414*.nc".format(channel),
            "gs://gcp-public-data-goes-16/ABI-L1b-RadM/2018/254/15/*RadM1*C{:02d}*s201825415*.nc".format(channel),
            "gs://gcp-public-data-goes-16/ABI-L1b-RadM/2018/254/16/*RadM1*C{:02d}*s201825416*.nc".format(channel),
        )]
    subdir = os.path.join(base_dir, "abi_l1b", "20180911_hurricane_florence_abi_l1b")
    os.makedirs(subdir, exist_ok=True)
    filenames = get_bucket_files(patterns, subdir, force=force, pattern_slice=frame_slice)

    actual_slice = frame_slice.indices(240)  # 240 max frames
    num_frames = int((actual_slice[1] - actual_slice[0]) / actual_slice[2])
    if len(filenames) != len(channels) * num_frames:
        raise RuntimeError("Not all files could be downloaded")
    return filenames
